# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np
from scipy import special

# some variables
sq2 = np.sqrt(2)
eps = np.finfo(np.float32).eps
l2p = np.log(2) + np.log(np.pi)


def joint_min(mu: np.ndarray, var: np.ndarray, with_derivatives: bool = False) -> np.ndarray:
    """
    Computes the probability of every given point to be the minimum
    based on the EPMGP[1] algorithm.
    [1] J. Cunningham, P. Hennig, and S. Lacoste-Julien.
    Gaussian probabilities and expectation propagation.
    under review. Preprint at arXiv, November 2011.

    :param mu: Mean value of each of the N points, dims (N,).
    :param var: Covariance matrix for all points, dims (N, N).
    :param with_derivatives: If True than also the gradients are computed.
    :returns: pmin distribution, dims (N,1).
    """

    logP = np.zeros(mu.shape)
    D = mu.shape[0]
    if with_derivatives:
        dlogPdMu = np.zeros((D, D))
        dlogPdSigma = np.zeros((D, int(0.5 * D * (D + 1))))
        dlogPdMudMu = np.zeros((D, D, D))
    for i in range(mu.shape[0]):

        # logP[k] ) self._min_factor(mu, var, 0)
        a = min_factor(mu, var, i)

        logP[i] = next(a)
        if with_derivatives:
            dlogPdMu[i, :] = next(a).T
            dlogPdMudMu[i, :, :] = next(a)
            dlogPdSigma[i, :] = next(a).T

    logP[np.isinf(logP)] = -500
    # re-normalize at the end, to smooth out numerical imbalances:
    logPold = logP
    Z = np.sum(np.exp(logPold))
    maxLogP = np.max(logP)
    s = maxLogP + np.log(np.sum(np.exp(logP - maxLogP)))
    s = maxLogP if np.isinf(s) else s

    logP = logP - s
    if not with_derivatives:
        return logP

    dlogPdMuold = dlogPdMu
    dlogPdSigmaold = dlogPdSigma
    dlogPdMudMuold = dlogPdMudMu
    # adjust derivatives, too. This is a bit tedious.
    Zm = sum(np.rot90((np.exp(logPold) * np.rot90(dlogPdMuold, 1)), 3)) / Z
    Zs = sum(np.rot90((np.exp(logPold) * np.rot90(dlogPdSigmaold, 1)), 3)) / Z

    dlogPdMu = dlogPdMuold - Zm
    dlogPdSigma = dlogPdSigmaold - Zs

    ff = np.einsum("ki,kj->kij", dlogPdMuold, dlogPdMuold)
    gg = np.einsum("kij,k->ij", dlogPdMudMuold + ff, np.exp(logPold)) / Z
    Zij = Zm.T * Zm
    adds = np.reshape(-gg + Zij, (1, D, D))
    dlogPdMudMu = dlogPdMudMuold + adds
    return logP, dlogPdMu, dlogPdSigma, dlogPdMudMu


def min_factor(Mu, Sigma, k, gamma=1):
    D = Mu.shape[0]
    logS = np.zeros((D - 1,))
    # mean time first moment
    MP = np.zeros((D - 1,))

    # precision, second moment
    P = np.zeros((D - 1,))

    M = np.copy(Mu)
    V = np.copy(Sigma)
    b = False
    d = np.NaN
    for count in range(50):
        diff = 0
        for i in range(D - 1):
            l = i if i < k else i + 1  # noqa: E741 to be consistent with paper notation
            try:
                M, V, P[i], MP[i], logS[i], d = lt_factor(k, l, M, V, MP[i], P[i], gamma)
            except Exception as e:
                raise

            if np.isnan(d):
                break
            diff += np.abs(d)
        if np.isnan(d):
            break
        if np.abs(diff) < 0.001:
            b = True
            break
    if np.isnan(d):
        logZ = -np.Infinity
        yield logZ
        dlogZdMu = np.zeros((D, 1))
        yield dlogZdMu

        dlogZdMudMu = np.zeros((D, D))
        yield dlogZdMudMu
        dlogZdSigma = np.zeros((int(0.5 * (D * (D + 1))), 1))
        yield dlogZdSigma
        mvmin = [Mu[k], Sigma[k, k]]
        yield mvmin
    else:
        # evaluate log Z:
        C = np.eye(D) / sq2
        C[k, :] = -1 / sq2
        C = np.delete(C, k, 1)

        R = np.sqrt(P.T) * C
        r = np.sum(MP.T * C, 1)
        mp_not_zero = np.where(MP != 0)
        mpm = MP[mp_not_zero] * MP[mp_not_zero] / P[mp_not_zero]
        mpm = sum(mpm)

        s = sum(logS)
        IRSR = np.eye(D - 1) + np.dot(np.dot(R.T, Sigma), R)
        rSr = np.dot(np.dot(r.T, Sigma), r)
        A = np.dot(R, np.linalg.solve(IRSR, R.T))

        A = 0.5 * (A.T + A)  # ensure symmetry.
        b = Mu + np.dot(Sigma, r)
        Ab = np.dot(A, b)
        try:
            cIRSR = np.linalg.cholesky(IRSR)
        except np.linalg.LinAlgError:
            try:
                cIRSR = np.linalg.cholesky(IRSR + 1e-10 * np.eye(IRSR.shape[0]))
            except np.linalg.LinAlgError:
                cIRSR = np.linalg.cholesky(IRSR + 1e-6 * np.eye(IRSR.shape[0]))
        dts = 2 * np.sum(np.log(np.diagonal(cIRSR)))
        logZ = 0.5 * (rSr - np.dot(b.T, Ab) - dts) + np.dot(Mu.T, r) + s - 0.5 * mpm
        yield logZ
        btA = np.dot(b.T, A)

        dlogZdMu = r - Ab
        yield dlogZdMu
        dlogZdMudMu = -A
        yield dlogZdMudMu
        dlogZdSigma = -A - 2 * np.outer(r, Ab.T) + np.outer(r, r.T) + np.outer(btA.T, Ab.T)
        dlogZdSigma2 = np.zeros_like(dlogZdSigma)
        np.fill_diagonal(dlogZdSigma2, np.diagonal(dlogZdSigma))
        dlogZdSigma = 0.5 * (dlogZdSigma + dlogZdSigma.T - dlogZdSigma2)
        dlogZdSigma = np.rot90(dlogZdSigma, k=2)[np.triu_indices(D)][::-1]
        yield dlogZdSigma


def lt_factor(s, l, M, V, mp, p, gamma):
    cVc = (V[l, l] - 2 * V[s, l] + V[s, s]) / 2.0
    Vc = (V[:, l] - V[:, s]) / sq2
    cM = (M[l] - M[s]) / sq2
    cVnic = np.max([cVc / (1 - p * cVc), 0])
    cmni = cM + cVnic * (p * cM - mp)
    z = cmni / np.sqrt(cVnic + 1e-25)
    if np.isnan(z):
        z = -np.inf
    e, lP, exit_flag = log_relative_gauss(z)
    if exit_flag == 0:
        alpha = e / np.sqrt(cVnic)
        # beta  = alpha * (alpha + cmni / cVnic);
        # r     = beta * cVnic / (1 - cVnic * beta);
        beta = alpha * (alpha * cVnic + cmni)
        r = beta / (1 - beta)
        # new message
        pnew = r / cVnic
        mpnew = r * (alpha + cmni / cVnic) + alpha

        # update terms
        dp = np.max([-p + eps, gamma * (pnew - p)])  # at worst, remove message
        dmp = np.max([-mp + eps, gamma * (mpnew - mp)])
        d = np.max([dmp, dp])  # for convergence measures

        pnew = p + dp
        mpnew = mp + dmp
        # project out to marginal
        Vnew = V - dp / (1 + dp * cVc) * np.outer(Vc, Vc)

        Mnew = M + (dmp - cM * dp) / (1 + dp * cVc) * Vc
        if np.any(np.isnan(Vnew)):
            raise Exception(
                "an error occurs while running expectation "
                "propagation in entropy search. "
                "Resulting variance contains NaN"
            )
        # % there is a problem here, when z is very large
        logS = lP - 0.5 * (np.log(beta) - np.log(pnew) - np.log(cVnic)) + (alpha * alpha) / (2 * beta) * cVnic

    elif exit_flag == -1:
        d = np.NAN
        Mnew = 0
        Vnew = 0
        pnew = 0
        mpnew = 0
        logS = -np.Infinity
    elif exit_flag == 1:
        d = 0
        # remove message from marginal:
        # new message
        pnew = 0
        mpnew = 0
        # update terms
        dp = -p  # at worst, remove message
        dmp = -mp
        d = max([dmp, dp])  # for convergence measures
        # project out to marginal
        Vnew = V - dp / (1 + dp * cVc) * (np.outer(Vc, Vc))
        Mnew = M + (dmp - cM * dp) / (1 + dp * cVc) * Vc
        logS = 0
    return Mnew, Vnew, pnew, mpnew, logS, d


def log_relative_gauss(z):
    """
    log_relative_gauss
    """
    if z < -6:
        return 1, -1.0e12, -1
    if z > 6:
        return 0, 0, 1
    else:
        logphi = -0.5 * (z * z + l2p)
        logPhi = np.log(0.5 * special.erfc(-z / sq2))
        e = np.exp(logphi - logPhi)
    return e, logPhi, 0



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import List

import numpy as np

from .user_function_result import UserFunctionResult


class LoopState(object):
    """
    Contains the state of the loop, which includes a history of all function evaluations
    """

    def __init__(self, initial_results: List[UserFunctionResult]) -> None:
        """
        :param initial_results: The function results from previous function evaluations
        """
        self.results = initial_results
        self.iteration = 0

    def update(self, results: List[UserFunctionResult]) -> None:
        """
        :param results: The latest function results since last update
        """
        if not results:
            raise ValueError("Cannot update state with empty result list.")

        self.iteration += 1
        self.results += results

    @property
    def X(self) -> np.ndarray:
        """
        :return: Function inputs for all function evaluations in a 2d array: number of points by input dimensions.
        """
        return np.array([result.X for result in self.results])

    @property
    def Y(self) -> np.ndarray:
        """
        :return: Function outputs for all function evaluations in a 2d array: number of points by output dimensions.
        """
        return np.array([result.Y for result in self.results])

    def __getattr__(self, item) -> np.array:
        """
        Overriding this method allows us to customise behaviour for accessing attributes. We use this to allow arbitrary
        fields in the loop state. These are usually extra outputs from the user function such as cost, constraint values
        etc. These fields are stored in each individual "UserFunctionResult" object in the "extra_outputs" dictionary

        :param item: The name of the item to acquire. Must match the key value in the "extra_outputs" dictionary in the
                     stored "UserFunctionResults" objects
        :return: The specified output for all function evaluations in a 2d array of size (n_points x n_dimensions)
        """
        # check key appears in all results objects
        is_valid = all([item in res.extra_outputs for res in self.results])
        if not is_valid:
            raise ValueError("{} not found in results object".format(item))
        return np.array([result.extra_outputs[item] for result in self.results])


def create_loop_state(x_init: np.ndarray, y_init: np.ndarray, **kwargs) -> LoopState:
    """
    Creates a loop state object using the provided data

    :param x_init: x values for initial function evaluations. Shape: (n_initial_points x n_input_dims)
    :param y_init: y values for initial function evaluations. Shape: (n_initial_points x n_output_dims)
    :param kwargs: extra outputs observed from a function evaluation. Shape: (n_initial_points x n_dims)
    """
    if x_init.shape[0] != y_init.shape[0]:
        error_message = "X and Y should have the same length. Actual length x_init {}, y_init {}".format(
            x_init.shape[0], y_init.shape[0]
        )
        raise ValueError(error_message)

    for key, value in kwargs.items():
        if value.shape[0] != x_init.shape[0]:
            raise ValueError(
                "Expected keyword argument {} to have length {} but actual length is {}".format(
                    key, x_init.shape[0], value.shape[0]
                )
            )

    initial_results = []
    for i in range(x_init.shape[0]):
        kwargs_dict = dict([(key, vals[i]) for key, vals in kwargs.items()])
        initial_results.append(UserFunctionResult(x_init[i], y_init[i], **kwargs_dict))

    return LoopState(initial_results)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import abc
import logging
import math

import numpy as np

from . import LoopState

_log = logging.getLogger(__name__)


class StoppingCondition(abc.ABC):
    """Chooses whether to stop the optimization based on the loop state"""

    def __and__(self, other: "StoppingCondition") -> "And":
        """
        Overloads self & other
        """
        return And(self, other)

    def __or__(self, other: "StoppingCondition") -> "Or":
        """
        Overloads self | other
        """
        return Or(self, other)

    @abc.abstractmethod
    def should_stop(self, loop_state: LoopState) -> bool:
        """
        :param loop_state: Object that contains current state of the loop
        :return: Whether to stop collecting new data
        """
        pass


class And(StoppingCondition):
    """
    Logical AND of two stopping conditions
    """

    def __init__(self, left: StoppingCondition, right: StoppingCondition):
        """
        :param left: One stopping condition in AND
        :param right: Another stopping condition in AND
        """
        self.left = left
        self.right = right

    def should_stop(self, loop_state: LoopState) -> bool:
        """
        Evaluate logical AND of two stopping conditions

        :param loop_state: Object that contains current state of the loop
        :return: Whether to stop collecting new data
        """
        return self.left.should_stop(loop_state) and self.right.should_stop(loop_state)


class Or(StoppingCondition):
    """
    Logical OR of two stopping conditions
    """

    def __init__(self, left: StoppingCondition, right: StoppingCondition):
        """
        :param left: One stopping condition in OR
        :param right: Another stopping condition in OR
        """
        self.left = left
        self.right = right

    def should_stop(self, loop_state: LoopState) -> bool:
        """
        Evaluate logical OR of two stopping conditions

        :param loop_state: Object that contains current state of the loop
        :return: Whether to stop collecting new data
        """
        return self.left.should_stop(loop_state) or self.right.should_stop(loop_state)


class FixedIterationsStoppingCondition(StoppingCondition):
    """Stops after a fixed number of iterations"""

    def __init__(self, i_max: int) -> None:
        """
        :param i_max: Maximum number of function
            observations within the loop, excluding initial points
        """
        self.i_max = i_max

    def should_stop(self, loop_state: LoopState) -> bool:
        """
        :param loop_state: Object that contains current state of the loop
        :return: True if maximum number of iterations has been reached
        """
        should_stop = loop_state.iteration >= self.i_max
        if should_stop is True:
            _log.info("Stopped after {} evaluations".format(self.i_max))
        return should_stop


class ConvergenceStoppingCondition(StoppingCondition):
    """Stops once we choose a point within eps of a previous
    point (with respect to euclidean norm). Close evaluations
    can suggest convergence of the optimization for problems
    with low observation noise.
    """

    def __init__(self, eps: float) -> None:
        """
        :param eps: minimum distance between
        two consecutive x's to keep running the model
        """
        self.eps = eps

    def should_stop(self, loop_state: LoopState) -> bool:
        """
        :param loop_state: Object that contains current state of the loop
        :return: True if the euclidean distance between the last two evaluations
                    is smaller than the specified eps.
        """
        if loop_state.iteration < 2:
            # less than 2 evaluations so cannot calculate distance
            return False
        should_stop = np.linalg.norm(loop_state.X[-1, :] - loop_state.X[-2, :]).item() <= self.eps
        if should_stop is True:
            _log.info("Stopped as consecutive evaluations are within {}".format(self.eps))
        return should_stop



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Tuple

import numpy as np

from ...core.acquisition import Acquisition
from ..methods import VanillaBayesianQuadrature
from .squared_correlation import SquaredCorrelation


class MutualInformation(Acquisition):
    r"""The mutual information between the integral value and integrand evaluations
    under a Gaussian process model.

    The mutual information is a monotonic transformation of the squared correlation, hence it
    yields the same acquisition policy under a standard Gaussian process model.

    .. math::
        a(x) = -0.5 \log(1-\rho^2(x))

    where :math:`\rho^2` is the squared correlation.

    .. seealso::
        :class:`emukit.quadrature.acquisitions.SquaredCorrelation`

    :param model: A vanilla Bayesian quadrature model.

    """

    def __init__(self, model: VanillaBayesianQuadrature):
        self.model = model
        self.rho2 = SquaredCorrelation(model)

    @property
    def has_gradients(self) -> bool:
        """Whether acquisition value has analytical gradient calculation available."""
        return True

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """Evaluates the acquisition function at x.

        :param x: The locations where to evaluate, shape (n_points, input_dim) .
        :return: The acquisition values at x, shape (n_points, 1).
        """
        rho2 = self.rho2.evaluate(x)
        mutual_information = -0.5 * np.log(1 - rho2)
        return mutual_information

    def evaluate_with_gradients(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Evaluate the acquisition function and its gradient.

        :param x: The locations where to evaluate, shape (n_points, input_dim) .
        :return: The acquisition values and corresponding gradients at x,
                 shapes (n_points, 1) and (n_points, input_dim)
        """
        # value
        mutual_information = self.evaluate(x)
        rho2, rho2_gradient = self.rho2.evaluate_with_gradients(x)

        # gradient
        mutual_information_gradient = (0.5 / (1 - rho2)) * rho2_gradient

        return mutual_information, mutual_information_gradient



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Tuple, Union

import numpy as np

from ...core.acquisition import Acquisition
from ...core.interfaces import IDifferentiable
from ..methods import WarpedBayesianQuadratureModel


class UncertaintySampling(Acquisition):
    r"""Uncertainty sampling.

    .. math::
        a(x) = \operatorname{var}(f(x)) p(x)^q

    where :math:`p(x)` is the density of the integration measure,
    :math:`\operatorname{var}(f(x))` is the predictive variance of the model at :math:`x`
    and :math:`q` is the ``measure_power`` parameter.

    :param model: A warped Bayesian quadrature model that has gradients.
    :param measure_power: The power :math:`q` of the measure. Default is 2.

    """

    def __init__(self, model: Union[WarpedBayesianQuadratureModel, IDifferentiable], measure_power: float = 2):
        self.model = model
        self._measure_power = measure_power

    @property
    def has_gradients(self) -> bool:
        """Whether acquisition value has analytical gradient calculation available."""
        return True

    def _evaluate(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Evaluate the predictive variances and the acquisition function.

        :param x: The locations where to evaluate, shape (n_points, input_dim) .
        :return: The acquisition values at x and unweighted variances, both of shape (n_points, 1).
        """
        variances = self.model.predict(x)[1]
        weights = self.model.measure.compute_density(x).reshape(variances.shape)
        return variances * weights**self._measure_power, variances

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """Evaluates the acquisition function at x.

        :param x: The locations where to evaluate, shape (n_points, input_dim) .
        :return: The acquisition values at x, shape (n_points, 1).
        """
        return self._evaluate(x)[0]

    def evaluate_with_gradients(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Evaluate the acquisition function and its gradient.

        :param x: The locations where to evaluate, shape (n_points, input_dim).
        :return: The acquisition values and corresponding gradients at x,
                 shapes (n_points, 1) and (n_points, input_dim)
        """
        p = self._measure_power
        variance_weighted, variance = self._evaluate(x)

        variance_gradient = self.model.get_prediction_gradients(x)[1]
        density = self.model.measure.compute_density(x)
        density_gradient = self.model.measure.compute_density_gradient(x)

        if p == 1:
            gradient_weighted = (density * variance_gradient.T).T + (variance[:, 0] * density_gradient.T).T
            return variance_weighted, gradient_weighted

        gradient_weighted = (density**p * variance_gradient.T).T + (
            p * (variance[:, 0] * density ** (p - 1)) * density_gradient.T
        ).T

        return variance_weighted, gradient_weighted



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .integrated_variance import IntegratedVarianceReduction  # noqa: F401
from .model_variance import ModelVariance  # noqa: F401



"""
This file implements an example of how to use Fabolas via the provided fmin interface to optimize the
hyperparameters of a SVM.
Instead of optimizing the real benchmark, we optimize a so-called surrogate benchmark instead which is much faster.
For more details look see th original Fabolas paper.

To run this example you need to install the new_benchmarks branch of HPOlib2:
    https://github.com/automl/HPOlib2/tree/new_benchmarks
"""
from emukit.core import ContinuousParameter, ParameterSpace
from emukit.examples.fabolas import fmin_fabolas

try:
    from hpolib.benchmarks.surrogates.svm import SurrogateSVM
except ImportError:
    raise ImportError("HPOLib is not installed. Please install it from: https://github.com/automl/HPOlib2/tree/master")

svm = SurrogateSVM()

l = []
for parameter in svm.get_configuration_space().get_hyperparameters():
    l.append(ContinuousParameter(parameter.name, parameter.lower, parameter.upper))

space = ParameterSpace(l)

s_min = 100
s_max = 50000


def wrapper(x, s):
    res = svm.objective_function(x, dataset_fraction=s / s_max)
    return res["function_value"], res["cost"]


res = fmin_fabolas(wrapper, space=space, s_min=s_min, s_max=s_max, n_iters=100, marginalize_hypers=False)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

# there are circular imports in this package, hence we can't just sort alphabetically
# isort: skip_file

from .expected_improvement import ExpectedImprovement, MultipointExpectedImprovement  # noqa: F401
from .negative_lower_confidence_bound import NegativeLowerConfidenceBound  # noqa: F401
from .probability_of_improvement import ProbabilityOfImprovement  # noqa: F401
from .probability_of_feasibility import ProbabilityOfFeasibility  # noqa: F401
from .entropy_search import EntropySearch  # noqa: F401
from .max_value_entropy_search import MaxValueEntropySearch  # noqa: F401
from .max_value_entropy_search import MUMBO



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .candidate_point_calculators import (  # noqa: F401
    CandidatePointCalculator,
    RandomSampling,
    SequentialPointCalculator,
)
from .loop_state import LoopState  # noqa: F401
from .model_updaters import FixedIntervalUpdater, ModelUpdater  # noqa: F401
from .outer_loop import OuterLoop  # noqa: F401
from .stopping_conditions import (  # noqa: F401
    ConvergenceStoppingCondition,
    FixedIterationsStoppingCondition,
    StoppingCondition,
)
from .user_function import UserFunction, UserFunctionWrapper  # noqa: F401
from .user_function_result import UserFunctionResult  # noqa: F401



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import List, Tuple

import numpy as np


class Parameter(object):
    def __init__(self, name: str):
        self.name = name

    def __str__(self):
        return f"<Parameter: {self.name}>"

    def __repr__(self):
        return f"Parameter({self.name})"

    @property
    def dimension(self) -> int:
        """
        Gives the dimension of the parameter.
        """
        return 1

    @property
    def model_parameters(self) -> List:
        """
        Gives the list of single dimensional model parameters the parameter corresponds to.
        """
        return [self]

    @property
    def bounds(self) -> List[Tuple]:
        """
        Returns bounds of the parameter in a form of list of tuples
        """
        raise NotImplemented

    def round(self, x: np.ndarray) -> np.ndarray:
        """
        Rounds the values of x to fit to the parameter domain, if needed.

        :param x: 2d array of values to be rounded.
        :returns: A 2d array of rounded values.
        """
        return x

    def check_in_domain(self, x: np.ndarray) -> bool:
        """
        Verifies that given values lie within the parameter's domain

        :param x: Value to be checked
        :return: A boolean value which indicates whether all points lie in the domain
        """
        raise NotImplementedError

    def sample_uniform(self, point_count: int) -> np.ndarray:
        """
        Generates multiple uniformly distributed random parameter points.

        :param point_count: number of data points to generate.
        :returns: Generated points with shape (point_count, num_features)
        """
        raise NotImplementedError



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import emcee


class McmcSampler(object):
    def __init__(self, space):
        """
        Creates an instance of the sampler.

        Parameters:
            space - variable space
        """
        self.space = space

    def get_samples(self, n_samples, log_p_function, burn_in_steps=50):
        """
        Generates samples.

        Parameters:
            n_samples - number of samples to generate
            log_p_function - a function that returns log density for a specific sample
            burn_in_steps - number of burn-in steps for sampling

        Returns a tuple of two lists: (samples, log_p_function values for samples)
        """

        raise NotImplementedError


class AffineInvariantEnsembleSampler(McmcSampler):
    def __init__(self, space):
        """
        Creates an instance of the affine invariant ensemble sampler.

        Parameters:
            space - variable space
        """
        super(AffineInvariantEnsembleSampler, self).__init__(space)

    def get_samples(self, n_samples, log_p_function, burn_in_steps=50, n_steps=100):
        """
        Generates samples.

        Parameters:
            n_samples - number of samples to generate
            log_p_function - a function that returns log density for a specific sample
            burn_in_steps - number of burn-in steps for sampling

        Returns a tuple of two array: (samples, log_p_function values for samples)
        """
        X_init = self.space.sample_uniform(n_samples)
        sampler = emcee.EnsembleSampler(n_samples, X_init.shape[1], log_p_function)

        # Burn-In
        state = list(sampler.run_mcmc(X_init, burn_in_steps))  # compatible with both emcee 2 and 3
        samples = state[0]
        samples_log = state[1]

        # MCMC Sampling
        state = list(sampler.run_mcmc(samples, n_steps))
        samples = state[0]
        samples_log = state[1]

        # make sure we have an array of shape (n samples, space input dim)
        if len(samples.shape) == 1:
            samples = samples.reshape(-1, 1)
        samples_log = samples_log.reshape(-1, 1)

        return samples, samples_log



"""The bounded Bayesian quadrature model is with square-root warping."""

from typing import Optional, Tuple

import numpy as np

from ..interfaces.base_gp import IBaseGaussianProcess
from ..kernels.quadrature_rbf import QuadratureRBFGaussianMeasure
from .warped_bq_model import WarpedBayesianQuadratureModel
from .warpings import SquareRootWarping


class BoundedBayesianQuadrature(WarpedBayesianQuadratureModel):
    r"""A warped Bayesian quadrature model that is upper bounded OR lower bounded by a constant.

    The integrand :math:`f(x)` is modeled as :math:`f(x) = f_* + \frac{1}{2} g(x)^2` for lower bounded functions,
    or as :math:`f(x) = f^* - \frac{1}{2}g(x)^2` for upper bounded functions.
    The constants :math:`f_*` and :math:`f^*` are the lower and upper bound respectively,
    and :math:`g` is a Gaussian process (GP).

    The process :math:`f` induced by the Gaussian process :math:`g` is non-Gaussian and not easy to integrate.
    In order to obtain an analytic estimator for the integral value, this class approximates the process :math:`f`
    by another Gaussian process :math:`\hat{f}` which is found by linearizing :math:`f`
    around the mean of :math:`g`. It is then possible to integrate :math:`\hat{f}` analytically.
    The approximate GP :math:`\hat{f}` is implemented in the predict methods in this class, and it is
    also used by :meth:`integrate`.

    .. seealso::
        :class:`emukit.quadrature.methods.warpings.SquareRootWarping`

    :param base_gp: The Gaussian process :math:`g`. Must use
           :class:`emukit.quadrature.kernels.QuadratureRBFGaussianMeasure` as kernel.
    :param X: The initial locations of integrand evaluations, shape (n_point, input_dim).
    :param Y: The values of the integrand at X, shape (n_points, 1).
    :param lower_bound: The lower bound  :math:`f_*` if the function is lower bounded.
    :param upper_bound: The upper bound :math:`f^*` if the function is lower bounded.

    :raises ValueError: If neither ``lower_bound`` nor ``upper_bound`` is given.
    :raises ValueError: If both ``lower_bound`` and ``upper_bound`` are given.
    :raises ValueError: If the quadrature kernel of ``base_gp`` is not of type :class:`QuadratureRBFGaussianMeasure`.

    """

    def __init__(
        self,
        base_gp: IBaseGaussianProcess,
        X: np.ndarray,
        Y: np.ndarray,
        lower_bound: Optional[float] = None,
        upper_bound: Optional[float] = None,
    ):
        if lower_bound is None and upper_bound is None:
            raise ValueError("Either a lower or an upper bound needs to be given. Currently neither is given.")
        if lower_bound is not None and upper_bound is not None:
            raise ValueError("Either a lower or an upper bound needs to be given. Currently both are given.")

        bound = lower_bound
        is_lower_bounded = True

        if lower_bound is None:
            bound = upper_bound
            is_lower_bounded = False

        # The integrate method is specific to QuadratureRBFGaussianMeasure, predict methods are only specific to
        # the approximation method used (Taylor expansion of the GP in this case).
        if not isinstance(base_gp.kern, QuadratureRBFGaussianMeasure):
            raise ValueError(
                f"{self.__class__.__name__} can only be used with QuadratureRBFGaussianMeasure kernel. "
                f"Instead {type(base_gp.kern)} is given."
            )

        super(BoundedBayesianQuadrature, self).__init__(
            base_gp=base_gp, warping=SquareRootWarping(offset=bound, is_inverted=not is_lower_bounded), X=X, Y=Y
        )

    @property
    def bound(self):
        """The bound :math:`f^*` or :math:`f_*` as defined in the model."""
        return self._warping.offset

    @property
    def is_lower_bounded(self):
        """``True`` if the model is lower bounded, ``False`` if it is upper bounded."""
        return not self._warping.is_inverted

    def predict_base(self, X_pred: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        mean_base, var_base = self.base_gp.predict(X_pred)

        mean_approx = self.transform(mean_base)
        var_approx = var_base * (mean_base**2)
        return mean_approx, var_approx, mean_base, var_base

    def predict_base_with_full_covariance(
        self, X_pred: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        mean_base, cov_base = self.base_gp.predict_with_full_covariance(X_pred)

        mean_approx = self.transform(mean_base)
        cov_approx = np.outer(mean_base, mean_base) * cov_base
        cov_approx = self.symmetrize_matrix(cov_approx)  # force symmetry for numerical stability
        return mean_approx, cov_approx, mean_base, cov_base

    def integrate(self) -> Tuple[float, float]:
        n_points, input_dim = self.X.shape

        # weights and kernel
        X = self.X / np.sqrt(2)  # this is equivalent to scaling the lengthscale with sqrt(2)
        K = self.base_gp.kern.K(X, X)
        weights = self.base_gp.graminv_residual()
        Weights_outer = np.outer(weights, weights)

        # kernel mean but with scaled lengthscale (multiplicative factor of 1/sqrt(2))
        X_means_vec = 0.5 * (self.X.T[:, :, None] + self.X.T[:, None, :]).reshape(input_dim, -1).T
        qK = self.base_gp.kern.qK(X_means_vec, scale_factor=1.0 / np.sqrt(2)).reshape(n_points, n_points)

        # integral mean
        integral_mean_second_term = 0.5 * np.sum(Weights_outer * qK * K)
        if self.is_lower_bounded:
            integral_mean = self.bound + integral_mean_second_term
        else:
            integral_mean = self.bound - integral_mean_second_term

        # integral variance
        # The integral variance is not needed for the WSABI loop as WSABI uses uncertainty sampling.
        # For completeness, the integral variance will need to be implemented at a later point.
        integral_variance = None
        return float(integral_mean), integral_variance

    def get_prediction_gradients(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        # predictions and gradients of base model
        mean_base, var_base = self.base_gp.predict(X)
        d_mean_dx_base, d_var_dx_base = self.base_gp.get_prediction_gradients(X)

        # gradient of mean
        d_mean_dx = (self.base_gp.kern.dK_dx1(X, self.X) @ self.base_gp.graminv_residual())[:, :, 0].T
        d_mean_dx = mean_base * d_mean_dx  # broadcasting  (n_points, 1) and (n_points, input_dim)

        # gradient of variance
        d_var_dx = (mean_base**2) * d_var_dx_base + (2 * var_base * mean_base) * d_mean_dx_base  # broadcasting

        # the gradient of the mean of the lower bounded model is the negative gradient of the upper bounded model.
        if not self.is_lower_bounded:
            d_mean_dx *= -1.0

        return d_mean_dx, d_var_dx



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Union

import numpy as np

from ...bayesian_optimization.acquisitions import ExpectedImprovement, ProbabilityOfFeasibility
from ...core.acquisition import Acquisition
from ...core.interfaces import IDifferentiable, IModel
from ...core.loop import FixedIntervalUpdater, OuterLoop, SequentialPointCalculator
from ...core.loop.loop_state import create_loop_state
from ...core.optimization import AcquisitionOptimizerBase, GradientAcquisitionOptimizer
from ...core.parameter_space import ParameterSpace
from ..acquisitions.log_acquisition import LogAcquisition
from ..local_penalization_calculator import LocalPenalizationPointCalculator


class UnknownConstraintBayesianOptimizationLoop(OuterLoop):
    def __init__(
        self,
        space: ParameterSpace,
        model_objective: Union[IModel, IDifferentiable],
        model_constraint: Union[IModel, IDifferentiable],
        acquisition: Acquisition = None,
        update_interval: int = 1,
        batch_size: int = 1,
    ):

        """
        Emukit class that implements a loop for building Bayesian optimization with an unknown constraint.
        For more information see:

        Michael A. Gelbart, Jasper Snoek, and Ryan P. Adams,
        Bayesian Optimization with Unknown Constraints,
        https://arxiv.org/pdf/1403.5607.pdf

        :param space: Input space where the optimization is carried out.
        :param model_objective: The model that approximates the underlying objective function
        :param model_constraint: The model that approximates the unknown constraints
        :param acquisition: The acquisition function for the objective function (default, EI).
        :param update_interval:  Number of iterations between optimization of model hyper-parameters. Defaults to 1.
        :param batch_size: How many points to evaluate in one iteration of the optimization loop. Defaults to 1.
        """

        if not np.all(np.isclose(model_objective.X, model_constraint.X)):
            raise ValueError(
                "Emukit currently only supports identical " "training inputs for the constrained and objective model"
            )

        if acquisition is None:
            acquisition = ExpectedImprovement(model_objective)

        acquisition_constraint = ProbabilityOfFeasibility(model_constraint)

        acquisition_constrained = acquisition * acquisition_constraint

        model_updater_objective = FixedIntervalUpdater(model_objective, update_interval)
        model_updater_constraint = FixedIntervalUpdater(
            model_constraint, update_interval, lambda state: state.Y_constraint
        )

        acquisition_optimizer = GradientAcquisitionOptimizer(space)
        if batch_size == 1:
            candidate_point_calculator = SequentialPointCalculator(acquisition_constrained, acquisition_optimizer)
        else:
            log_acquisition = LogAcquisition(acquisition_constrained)
            candidate_point_calculator = LocalPenalizationPointCalculator(
                log_acquisition, acquisition_optimizer, model_objective, space, batch_size
            )

        loop_state = create_loop_state(model_objective.X, model_objective.Y, Y_constraint=model_constraint.Y)

        super(UnknownConstraintBayesianOptimizationLoop, self).__init__(
            candidate_point_calculator, [model_updater_objective, model_updater_constraint], loop_state
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import numpy as np

from ...core.interfaces.models import IModel

try:
    from sklearn.ensemble import RandomForestRegressor
except ImportError:
    raise ImportError(
        """
        This module is missing required dependencies. Try running:

        pip install scikit-learn

        Refer to http://scikit-learn.org/stable/install.html for further information.
    """
    )


class RandomForest(IModel):
    def __init__(
        self,
        X_init: np.ndarray,
        Y_init: np.ndarray,
        num_trees: int = 30,
        do_bootstrapping: bool = True,
        seed: int = None,
    ) -> None:
        """
        Interface to random forests for Bayesian optimization based on Scikit Learn package.

        Dependencies:
            Scikit Learn (http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestRegressor.html)

        :param X_init: Initial input data points to train the model
        :param Y_init: Initial target values
        :param num_trees: Specifies the number of trees to build the random forest
        :param do_bootstrapping: Defines if we use boostrapping for the individual trees or not
        :param seed: Used to seed the random number generator for the random forest (None means random seed)
        """
        super().__init__()

        self._X = X_init
        self._Y = Y_init

        self.rf = RandomForestRegressor(n_estimators=num_trees, bootstrap=do_bootstrapping, random_state=seed)
        self.rf.fit(X_init, Y_init[:, 0])

    def predict(self, X: np.ndarray) -> np.ndarray:
        """
        Predict values for given points

        :param X: points to run prediction for
        :return: Tuple of mean and variance which are 2d arrays of shape (n_points x n_outputs)
        """

        preds = []
        for estimator in self.rf.estimators_:
            pred = estimator.predict(X)
            preds.append(pred)
        mean = np.array(preds).mean(axis=0)[:, None]
        var = np.array(preds).var(axis=0)[:, None]
        return mean, var

    def set_data(self, X: np.ndarray, Y: np.ndarray) -> None:
        """
        Sets training data in model

        :param X: new points
        :param Y: function values at new points X
        """
        self._X = X
        self._Y = Y

        self.rf.fit(X, Y[:, 0])

    def optimize(self) -> None:
        pass

    def get_f_minimum(self):
        """Return the minimum value that has been observed"""
        return np.min(self._Y)

    @property
    def X(self):
        return self._X

    @property
    def Y(self):
        return self._Y



"""The product Matern32 kernel embeddings."""

from typing import Union

import numpy as np

from ...quadrature.interfaces.standard_kernels import IProductMatern32
from ..measures import IntegrationMeasure, LebesgueMeasure
from .quadrature_kernels import LebesgueEmbedding, QuadratureProductKernel


class QuadratureProductMatern32(QuadratureProductKernel):
    r"""Base class for a product Matern32 kernel augmented with integrability.

    The kernel is of the form :math:`k(x, x') = \sigma^2 \prod_{i=1}^d k_i(x, x')` where

    .. math::
        k_i(x, x') = (1 + \sqrt{3}r_i ) e^{-\sqrt{3} r_i}.

    Above, :math:`d` is the input dimensionality, :math:`r_i =\frac{|x_i - x'_i|}{\lambda_i}`,
    is the scaled distance, :math:`\sigma^2` is the ``variance`` property and :math:`\lambda_i`
    is the :math:`i` th element of the ``lengthscales`` property.

    .. note::
        This class is compatible with the standard kernel :class:`IProductMatern32`.
        Each subclass of this class implements an embedding w.r.t. a specific integration measure.

    .. seealso::
       * :class:`emukit.quadrature.interfaces.IProductMatern32`
       * :class:`emukit.quadrature.kernels.QuadratureProductKernel`

    :param matern_kernel: The standard EmuKit product Matern32 kernel.
    :param measure: The integration measure.
    :param variable_names: The (variable) name(s) of the integral.

    """

    def __init__(
        self,
        matern_kernel: IProductMatern32,
        measure: IntegrationMeasure,
        variable_names: str,
    ) -> None:
        super().__init__(kern=matern_kernel, measure=measure, variable_names=variable_names)

    @property
    def nu(self) -> float:
        """The smoothness parameter of the kernel."""
        return self.kern.nu

    @property
    def lengthscales(self) -> np.ndarray:
        r"""The lengthscales :math:`\lambda` of the kernel."""
        return self.kern.lengthscales

    @property
    def variance(self) -> float:
        r"""The scale :math:`\sigma^2` of the kernel."""
        return self.kern.variance


class QuadratureProductMatern32LebesgueMeasure(QuadratureProductMatern32, LebesgueEmbedding):
    """A product Matern32 kernel augmented with integrability w.r.t. the standard Lebesgue measure.

    .. seealso::
       * :class:`emukit.quadrature.interfaces.IProductMatern32`
       * :class:`emukit.quadrature.kernels.QuadratureProductMatern32`
       * :class:`emukit.quadrature.measures.LebesgueMeasure`

    :param matern_kernel: The standard EmuKit product Matern32 kernel.
    :param measure: The Lebesgue measure.
    :param variable_names: The (variable) name(s) of the integral.

    """

    def __init__(self, matern_kernel: IProductMatern32, measure: LebesgueMeasure, variable_names: str = "") -> None:
        super().__init__(matern_kernel=matern_kernel, measure=measure, variable_names=variable_names)

    def _scale(self, z: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        return self.variance * z

    def _get_univariate_parameters(self, dim: int) -> dict:
        return {
            "domain": self.measure.domain.bounds[dim],
            "lengthscale": self.lengthscales[dim],
            "normalize": self.measure.is_normalized,
        }

    def _qK_1d(self, x: np.ndarray, **parameters) -> np.ndarray:
        a, b = parameters["domain"]
        lengthscale = parameters["lengthscale"]
        normalization = 1 / (b - a) if parameters["normalize"] else 1.0
        s3 = np.sqrt(3.0)
        first_term = 4.0 * lengthscale / s3
        second_term = -np.exp(s3 * (x - b) / lengthscale) * (b + 2.0 * lengthscale / s3 - x)
        third_term = -np.exp(s3 * (a - x) / lengthscale) * (x + 2.0 * lengthscale / s3 - a)
        return (first_term + second_term + third_term) * normalization

    def _qKq_1d(self, **parameters) -> float:
        a, b = parameters["domain"]
        lengthscale = parameters["lengthscale"]
        normalization = 1 / (b - a) if parameters["normalize"] else 1.0
        c = np.sqrt(3.0) * (b - a)
        qKq = (
            2.0 * lengthscale / 3.0 * (2.0 * c - 3.0 * lengthscale + np.exp(-c / lengthscale) * (c + 3.0 * lengthscale))
        )
        return float(qKq) * normalization**2

    def _dqK_dx_1d(self, x: np.ndarray, **parameters) -> np.ndarray:
        a, b = parameters["domain"]
        lengthscale = parameters["lengthscale"]
        normalization = 1 / (b - a) if parameters["normalize"] else 1.0
        s3 = np.sqrt(3)
        exp_term_b = np.exp(s3 * (x - b) / lengthscale)
        exp_term_a = np.exp(s3 * (a - x) / lengthscale)
        first_term = exp_term_b * (-1 + (s3 / lengthscale) * (x - b))
        second_term = exp_term_a * (+1 - (s3 / lengthscale) * (a - x))
        return (first_term + second_term) * normalization



from typing import Tuple

import numpy as np

from ...core import ContinuousParameter, InformationSourceParameter, ParameterSpace
from ...core.loop.user_function import MultiSourceFunctionWrapper


def multi_fidelity_currin_function() -> Tuple[MultiSourceFunctionWrapper, ParameterSpace]:
    r"""

    High fidelity function is given by:

    .. math::
        f_{high}(\mathbf{x}) = \left[ 1 - \exp \left(-\frac{1}{2x_2}\right) \right]
        \frac{2300x_1^3 + 1900x_1^2 + 2092x_1 + 60}{100x_1^3+500x_1^2 + 4x_1 + 20}

    Low fidelity function given by:

    .. math::
        f_{low}(\mathbf{x}) = \frac{1}{4} \left[ f_{high}(x_1 + 0.05, x_2 + 0.05) + f_{high}(x_1 + 0.05, \max (0, x_2 - 0.05)) \\
        +  f_{high}(x_1 - 0.05, x_2 + 0.05) + f_{high}\left(x_1 - 0.05, \max \left(0, x_2 - 0.05\right)\right) \right]

    Input domain:

    .. math::
        \mathbf{x}_i \in [0, 1]

    Reference: https://www.sfu.ca/~ssurjano/curretal88exp.html
    """

    def high(x):
        x1 = x[:, 0]
        x2 = x[:, 1]
        return (
            1
            - np.exp(-0.5 / x2)
            * ((2300 * x1**3 + 1900 * x1**2 + 2092 * x1 + 60) / (100 * x1**3 + 500 * x1**2 + 4 * x1 + 20))
        )[:, None]

    def low(x):
        return (
            0.25 * high(np.stack([x[:, 0] + 0.05, x[:, 1] + 0.05], axis=1))
            + 0.25 * high(np.stack([x[:, 0] + 0.05, np.maximum(0, x[:, 1] - 0.05)], axis=1))
            + 0.25 * high(np.stack([x[:, 0] - 0.05, x[:, 1] + 0.05], axis=1))
            + 0.25 * high(np.stack([x[:, 0] - 0.05, np.maximum(0, x[:, 1] - 0.05)], axis=1))
        )

    space = ParameterSpace(
        [ContinuousParameter("x1", 0, 1), ContinuousParameter("x2", 0, 1), InformationSourceParameter(2)]
    )
    return MultiSourceFunctionWrapper([low, high]), space



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import List, Tuple

import numpy as np
from scipy.integrate import dblquad, quad


def univariate_approximate_ground_truth_integral(func, integral_bounds: Tuple[float, float]):
    """Estimate the 1D ground truth integral.

    :param func: A univariate function.
    :param integral_bounds: Bounds of the integral.
    :returns: The integral estimate (output of scipy.integrate.quad).
    """
    lower_bound = integral_bounds[0]
    upper_bound = integral_bounds[1]
    return quad(func, lower_bound, upper_bound)


def bivariate_approximate_ground_truth_integral(func, integral_bounds: List[Tuple[float, float]]):
    """Estimate the 2D ground truth integral.

    :param func: A bivariate function.
    :param integral_bounds: Bounds of the integral.
    :returns: The integral estimate (output of scipy.integrate.dblquad).
    """

    def func_dblquad(x, y):
        z = np.array([x, y])
        z = np.reshape(z, [1, 2])
        return func(z)

    lower_bound_x = integral_bounds[0][0]
    upper_bound_x = integral_bounds[0][1]

    lower_bound_y = integral_bounds[1][0]
    upper_bound_y = integral_bounds[1][1]

    return dblquad(
        func=func_dblquad, a=lower_bound_x, b=upper_bound_x, gfun=lambda x: lower_bound_y, hfun=lambda x: upper_bound_y
    )



from typing import Tuple

import numpy as np

from ...core import ContinuousParameter, InformationSourceParameter, ParameterSpace
from ...core.loop.user_function import MultiSourceFunctionWrapper
from ..branin import _branin


def multi_fidelity_branin_function() -> Tuple[MultiSourceFunctionWrapper, ParameterSpace]:
    r"""
    Two-dimensional Branin, often used as an optimization benchmark but here modified to be used as a multi-fidelity
    benchmark

    Based on: https://www.sfu.ca/~ssurjano/branin.html

    High fidelity given by:

    .. math::
        f(\mathbf{x}) = (x_2 - b x_1 ^ 2 + c x_1 - r) ^ 2 + s(1 - t) \cos(x_1) + s

    Middle fidelity given by:

    .. math::
        f_{m}(\mathbf{x}) = \sqrt{f_{high}(\mathbf{x} - 2)} + \frac{2(x_1 - 0.5) - 3(3x_2 - 1) - 1}{100}

    Low fidelity given by:

    .. math::
        f_{m}(1.2(\mathbf{x} + 2)) - \frac{3 x_2 + 1}{100}


    where:

    .. math::
        b = 5.1 / (4 \pi ^ 2)

        c = 5 /\pi

        r = 6

        s = 10

        t = 1 / (8\pi)
    """

    def branin_medium_fidelity(x):
        x1 = x[:, 0]
        x2 = x[:, 1]
        result = (10.0 * np.sqrt(_branin(x - 2.0)[:, 0]) + 2.0 * (x1 - 0.5) - 3.0 * (3.0 * x2 - 1.0) - 1.0) / 100.0
        return result[:, None]

    def branin_low_fidelity(x):
        x2 = x[:, 1]
        result = (branin_medium_fidelity(1.2 * (x + 2.0))[:, 0] * 100.0 - 3.0 * x2 + 1.0) / 100.0
        return result[:, None]

    parameter_space = ParameterSpace(
        [ContinuousParameter("x1", -5, 10), ContinuousParameter("x2", 0, 15), InformationSourceParameter(3)]
    )

    branin_high = lambda x: _branin(x) / 100
    return MultiSourceFunctionWrapper([branin_low_fidelity, branin_medium_fidelity, branin_high]), parameter_space



"""Warpings as used by warped Bayesian quadrature models."""

import abc
from typing import Optional

import numpy as np


class Warping(abc.ABC):
    """Base class for a warping as used by a warped Bayesian quadrature model.

    .. seealso::
        * :class:`emukit.quadrature.methods.WarpedBayesianQuadratureModel`
        * :class:`emukit.quadrature.methods.warpings.IdentityWarping`
        * :class:`emukit.quadrature.methods.warpings.SquareRootWarping`

    """

    @abc.abstractmethod
    def transform(self, Y: np.ndarray) -> np.ndarray:
        """Transform from base-GP to integrand.

        :param Y: Function values of latent function, shape (n_points, 1).
        :return: Transformed values, shape (n_points, 1).
        """
        raise NotImplementedError

    @abc.abstractmethod
    def inverse_transform(self, Y: np.ndarray) -> np.ndarray:
        """Transform from integrand to base-GP.

        :param Y: Function values of integrand, shape (n_points, 1).
        :return: Transformed values, shape (n_points, 1).
        """
        raise NotImplementedError

    def update_parameters(self, **new_parameters) -> None:
        """Update the warping parameters.

        :param new_parameters: Contains the parameter names as keys with the new values.
                               An empty dictionary will do nothing.
        """
        for parameter, new_value in new_parameters.items():
            setattr(self, parameter, new_value)


class IdentityWarping(Warping):
    """The identity warping

    .. math::
        w(y) = y.

    """

    def transform(self, Y: np.ndarray) -> np.ndarray:
        return Y

    def inverse_transform(self, Y: np.ndarray) -> np.ndarray:
        return Y


class SquareRootWarping(Warping):
    r"""The square root warping.

    .. math::
        w(y)=\begin{cases}
                  c + \frac{1}{2}y^2 & \text{is_inverted is False (default)}\\
                  c - \frac{1}{2}y^2 &\text{otherwise}
              \end{cases},

    where :math:`c` is a constant.

    :param offset: The offset :math:`c` of the warping.
    :param is_inverted: Inverts the warping if ``True``. Default is ``False``.

    """

    def __init__(self, offset: float, is_inverted: Optional[bool] = False):
        self.offset = offset
        self.is_inverted = is_inverted

    def transform(self, Y: np.ndarray) -> np.ndarray:
        if self.is_inverted:
            return self.offset - 0.5 * (Y * Y)
        else:
            return self.offset + 0.5 * (Y * Y)

    def inverse_transform(self, Y: np.ndarray) -> np.ndarray:
        if self.is_inverted:
            return np.sqrt(2.0 * (self.offset - Y))
        else:
            return np.sqrt(2.0 * (Y - self.offset))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import logging

import numpy as np

from ...core.acquisition import Acquisition
from ...core.interfaces import IDifferentiable, IModel
from ...core.loop import FixedIntervalUpdater, OuterLoop, SequentialPointCalculator
from ...core.loop.loop_state import LoopState, create_loop_state
from ...core.optimization import AcquisitionOptimizerBase, GradientAcquisitionOptimizer
from ...core.parameter_space import ParameterSpace
from ..acquisitions import ExpectedImprovement
from ..acquisitions.log_acquisition import LogAcquisition
from ..local_penalization_calculator import LocalPenalizationPointCalculator

_log = logging.getLogger(__name__)


class BayesianOptimizationLoop(OuterLoop):
    def __init__(
        self,
        space: ParameterSpace,
        model: IModel,
        acquisition: Acquisition = None,
        update_interval: int = 1,
        batch_size: int = 1,
        acquisition_optimizer: AcquisitionOptimizerBase = None,
    ):

        """
        Emukit class that implement a loop for building modular Bayesian optimization

        :param space: Input space where the optimization is carried out.
        :param model: The model that approximates the underlying function
        :param acquisition: The acquisition function that will be used to collect new points (default, EI). If batch
                            size is greater than one, this acquisition must output positive values only.
        :param update_interval: Number of iterations between optimization of model hyper-parameters. Defaults to 1.
        :param batch_size: How many points to evaluate in one iteration of the optimization loop. Defaults to 1.
        :param acquisition_optimizer: Optimizer selecting next evaluation points
                                      by maximizing acquisition.
                                      Gradient based optimizer is used if None.
                                      Defaults to None.
        """

        self.model = model

        if acquisition is None:
            acquisition = ExpectedImprovement(model)

        model_updaters = FixedIntervalUpdater(model, update_interval)

        if acquisition_optimizer is None:
            acquisition_optimizer = GradientAcquisitionOptimizer(space)
        if batch_size == 1:
            _log.info("Batch size is 1, using SequentialPointCalculator")
            candidate_point_calculator = SequentialPointCalculator(acquisition, acquisition_optimizer)
        else:
            _log.info("Batch size is " + str(batch_size) + ", using LocalPenalizationPointCalculator")
            log_acquisition = LogAcquisition(acquisition)
            candidate_point_calculator = LocalPenalizationPointCalculator(
                log_acquisition, acquisition_optimizer, model, space, batch_size
            )

        loop_state = create_loop_state(model.X, model.Y)

        super().__init__(candidate_point_calculator, model_updaters, loop_state)

    def get_results(self):
        return BayesianOptimizationResults(self.loop_state)


class BayesianOptimizationResults:
    def __init__(self, loop_state: LoopState):

        """
        Emukit class that takes as input the loop state and computes some results.

        :param loop_state: The loop state it its current form. Currently it only contains X and Y.
        """

        self.minimum_location = loop_state.X[np.argmin(loop_state.Y), :]
        self.minimum_value = np.min(loop_state.Y)
        self.best_found_value_per_iteration = np.minimum.accumulate(loop_state.Y).flatten()



from typing import Tuple

import numpy as np

from ..acquisition import Acquisition
from ..interfaces import IDifferentiable, IModel


def acquisition_per_expected_cost(acquisition: Acquisition, cost_model: IModel, min_cost: float = 1e-4) -> Acquisition:
    """
    Creates an acquisition function that is the original acquisition scaled by the expected value of the evaluation
    cost of the user function.

    :param acquisition: Base acquisition function
    :param cost_model: Model of the evaluation cost. Should return positive values only.
    :return: Scaled acquisition function
    """
    return acquisition / CostAcquisition(cost_model, min_cost)


class CostAcquisition(Acquisition):
    """
    Acquisition that simply returns the expected value from the cost model
    """

    def __init__(self, cost_model: IModel, min_cost: float = 1e-4):
        """
        :param cost_model: Model of cost. Should return only positive predictions
        :param min_cost: A minimum value for the cost. The cost model prediction will be clipped to this value if
                         required
        """
        self.cost_model = cost_model
        self.min_cost = min_cost

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluate acquisition function

        :param x: A numpy array of shape (n_inputs x n_input_features) containing input locations at which to evaluate
                  the cost
        :return: Value of expected cost at input locations
        """
        return np.maximum(self.cost_model.predict(x)[0], self.min_cost)

    @property
    def has_gradients(self) -> bool:
        """
        Whether gradients of the cost function with respect to input location are available
        """
        return isinstance(self.cost_model, IDifferentiable)

    def evaluate_with_gradients(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """

        :param x: A numpy array of shape (n_inputs x n_input_features) containing input locations at which to evaluate
                  the cost
        :return: Tuple of numpy arrays: (cost value, cost gradients)
        """
        expected_cost = self.cost_model.predict(x)[0]
        is_below_min_cost = (expected_cost < self.min_cost).flatten()
        grad = self.cost_model.get_prediction_gradients(x)[0]
        grad[is_below_min_cost, :] = 0
        return np.maximum(expected_cost, self.min_cost), grad



"""The Lebesgue measure."""

import numpy as np

from ...core.optimization.context_manager import ContextManager
from ..typing import BoundsType
from .domain import BoxDomain
from .integration_measure import IntegrationMeasure


class LebesgueMeasure(IntegrationMeasure):
    r"""The Lebesgue measure.

    The Lebesgue measure has density

    .. math::
        p(x)=\begin{cases} \hat{p} & x\in\text{domain}\\0 &\text{otherwise}\end{cases}.

    The value :math:`\hat{p} = 1` if the parameter ``normalized`` is set to ``False``
    and  :math:`\hat{p} = |\text{domain}|^{-1}` otherwise, where :math:`|\text{domain}|`
    is the volume (un-normalized Lebesgue measure) of the domain.

    :param domain: The Box domain.
    :param normalized: Weather the Lebesgue measure is normalized.

    :raises NumericalPrecisionError: If ``normalized=True`` this excetion can be raised if the volume of the domain
                                     is so small that it is numerically zero or even negative.

    """

    def __init__(self, domain: BoxDomain, normalized: bool = False):
        super().__init__(domain=domain, name="LebesgueMeasure")

        density = 1.0
        if normalized:
            differences = np.array([x[1] - x[0] for x in self.domain.bounds])
            volume = np.prod(differences)

            if volume <= 0:
                raise NumericalPrecisionError(
                    "Domain volume of uniform measure is not positive. Its value is {}.".format(volume)
                )
            density = float(1.0 / volume)

        self.density = density
        self.is_normalized = normalized
        self._input_dim = self.domain.dim

    @property
    def input_dim(self):
        return self._input_dim

    @property
    def can_sample(self) -> bool:
        return True

    def compute_density(self, x: np.ndarray) -> np.ndarray:
        # Compute density: (i) check if points are inside the box. (ii) multiply this bool value with density.
        bounds_lower = np.array([b[0] for b in self.domain.bounds])
        bounds_upper = np.array([b[1] for b in self.domain.bounds])
        inside_lower = 1 - (x < bounds_lower)  # contains 1 if element in x is above its lower bound, 0 otherwise
        inside_upper = 1 - (x > bounds_upper)  # contains 1 if element in x is below its upper bound, 0 otherwise

        # Contains True if element in x is inside box, False otherwise.
        # The returned array thus contains self._density for each point inside the box and 0 otherwise.
        inside_upper_lower = (inside_lower * inside_upper).sum(axis=1) == x.shape[1]
        return inside_upper_lower * self.density

    def compute_density_gradient(self, x: np.ndarray) -> np.ndarray:
        return np.zeros(x.shape)

    def reasonable_box(self) -> BoundsType:
        return self.domain.bounds

    def sample(self, num_samples: int, context_manager: ContextManager = None) -> np.ndarray:
        bounds = np.asarray(self.domain.bounds)

        samples = np.random.rand(num_samples, self.input_dim) * (bounds[:, 1] - bounds[:, 0]) + bounds[:, 0]

        if context_manager is not None:
            samples[:, context_manager.context_idxs] = context_manager.context_values

        return samples

    @classmethod
    def from_bounds(cls, bounds: BoundsType, normalized: bool = False):
        """Creates and instance of this class from integration bounds.

        .. seealso::
            * :class:`emukit.quadrature.measures.BoxDomain`

        :param bounds: List of :math:`d` tuples :math:`[(a_1, b_1), (a_2, b_2), \dots, (a_d, b_d)]`,
                       where :math:`d` is the dimensionality of the domain and the tuple :math:`(a_i, b_i)`
                       contains the lower and upper bound of dimension :math:`i` defining the box domain.
        :param normalized: Weather the Lebesgue measure is normalized.
        :return: An instance of LebesgueMeasure.

        """
        domain = BoxDomain(name="", bounds=bounds)
        return cls(domain=domain, normalized=normalized)


class NumericalPrecisionError(Exception):
    pass



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from enum import Enum

import numpy as np
from GPy.kern import Matern52
from GPy.models import GPRegression

from ...bayesian_optimization.acquisitions import (
    ExpectedImprovement,
    NegativeLowerConfidenceBound,
    ProbabilityOfImprovement,
)
from ...bayesian_optimization.loops import BayesianOptimizationLoop
from ...core.loop import FixedIterationsStoppingCondition, UserFunction
from ...core.parameter_space import ParameterSpace
from ...model_wrappers.gpy_model_wrappers import GPyModelWrapper
from .enums import AcquisitionType


class OptimizerType(Enum):
    LBFGS = 1


class GPBayesianOptimization(BayesianOptimizationLoop):
    def __init__(
        self,
        variables_list: list,
        X: np.array,
        Y: np.array,
        noiseless: bool = False,
        acquisition_type: AcquisitionType = AcquisitionType.EI,
        normalize_Y: bool = True,
        acquisition_optimizer_type: OptimizerType = OptimizerType.LBFGS,
        model_update_interval: int = int(1),
        batch_size: int = 1,
    ) -> None:

        """
        Generic class to run Bayesian optimization with GPyRegression model.

        Dependencies:
            GPy (https://github.com/SheffieldML/GPy)

        :param variables_list: list containing the definition of the variables of the input space.
        :param noiseless:  determines whether the objective function is noisy or not
        :param X: initial input values where the objective has been evaluated.
        :param Y: initial output values where the objective has been evaluated.
        :param acquisition_type: type of acquisition to use during optimization.
            - EI: Expected improvement
            - PI: Probability of improvement
            - NLCB: Negative lower confidence bound
        :param normalize_Y: whether the outputs of Y are normalized in the model.
        :param acquisition_optimizer_type: selects the type of optimizer of the acquisition.
            - LBFGS: uses L-BFGS with multiple initializations.
        :param model_update_interval: interval of interactions in which the model is updated.
        :param batch_size: How many points to evaluate in one iteration of the optimization loop. Defaults to 1.
        """

        self.variables_list = variables_list
        self.noiseless = noiseless
        self.X = X
        self.Y = Y
        self.acquisition_type = acquisition_type
        self.normalize_Y = normalize_Y
        self.acquisition_optimizer_type = acquisition_optimizer_type
        self.model_update_interval = model_update_interval

        # 1. Crete the internal object to handle the input space
        self.space = ParameterSpace(variables_list)

        # 2. Select the model to use in the optimization
        self._model_chooser()

        # 3. Select the acquisition function
        self._acquisition_chooser()

        super(GPBayesianOptimization, self).__init__(
            model=self.model, space=self.space, acquisition=self.acquisition, batch_size=batch_size
        )

    def _model_chooser(self):
        """Initialize the model used for the optimization"""
        kernel = Matern52(len(self.variables_list), variance=1.0, ARD=False)
        gpmodel = GPRegression(self.X, self.Y, kernel)
        gpmodel.optimize()
        self.model = GPyModelWrapper(gpmodel)
        if self.noiseless:
            gpmodel.Gaussian_noise.constrain_fixed(0.001)
        self.model = GPyModelWrapper(gpmodel)

    def _acquisition_chooser(self):
        """Select the acquisition function used in the optimization"""
        if self.acquisition_type is AcquisitionType.EI:
            self.acquisition = ExpectedImprovement(self.model)
        elif self.acquisition_type is AcquisitionType.PI:
            self.acquisition = ProbabilityOfImprovement(self.model)
        elif self.acquisition_type is AcquisitionType.NLCB:
            self.acquisition = NegativeLowerConfidenceBound(self.model)

    def suggest_new_locations(self):
        """Returns one or a batch of locations without evaluating the objective"""
        return self.candidate_point_calculator.compute_next_points(self.loop_state)

    def run_optimization(self, user_function: UserFunction, num_iterations: int) -> None:
        """
        :param user_function: The function that we want to optimize
        :param num_iterations: The number of iterations to run the Bayesian optimization loop.
        """
        self.run_loop(user_function, FixedIterationsStoppingCondition(num_iterations))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Callable, Union

import numpy as np
import scipy
from scipy.integrate import simps
from scipy.optimize import bisect
from scipy.stats import norm

from ...core import InformationSourceParameter
from ...core.acquisition import Acquisition
from ...core.initial_designs import RandomDesign
from ...core.interfaces import IModel
from ...core.parameter_space import ParameterSpace
from ..interfaces import IEntropySearchModel


class MaxValueEntropySearch(Acquisition):
    def __init__(
        self,
        model: Union[IModel, IEntropySearchModel],
        space: ParameterSpace,
        num_samples: int = 10,
        grid_size: int = 5000,
    ) -> None:

        """
        MES acquisition function approximates the distribution of the value at the global
        minimum and tries to decrease its entropy. See this paper for more details:
        Z. Wang, S. Jegelka
        Max-value Entropy Search for Efficient Bayesian Optimization
        ICML 2017

        :param model: GP model to compute the distribution of the minimum dubbed pmin.
        :param space: Domain space which we need for the sampling of the representer points
        :param num_samples: integer determining how many samples to draw of the minimum (does not need to be large)
        :param grid_size: number of random locations in grid used to fit the gumbel distribution and approximately generate
        the samples of the minimum (recommend scaling with problem dimension, i.e. 10000*d)
        """
        super().__init__()

        if not isinstance(model, IEntropySearchModel):
            raise RuntimeError("Model is not supported for MES")

        self.model = model
        self.space = space
        self.num_samples = num_samples
        self.grid_size = grid_size

        # Initialize parameters to lazily compute them once needed
        self.mins = None

    def update_parameters(self):
        """
        MES requires acces to a sample of possible minimum values y* of the objective function.
        To build this sample we approximate the empirical c.d.f of Pr(y*<y) with a Gumbel(a,b) distribution.
        This Gumbel distribution can then be easily sampled to yield approximate samples of y*

        This needs to be called once at the start of each BO step.
        """

        # First we generate a random grid of locations at which to fit the Gumbel distribution
        random_design = RandomDesign(self.space)
        grid = random_design.get_samples(self.grid_size)
        # also add the locations already queried in the previous BO steps
        grid = np.vstack([self.model.X, grid])
        # Get GP posterior at these points
        fmean, fvar = self.model.predict(grid)
        fsd = np.sqrt(fvar)

        # fit Gumbel distriubtion
        a, b = _fit_gumbel(fmean, fsd)

        # sample K times from this Gumbel distribution using the inverse probability integral transform,
        # i.e. given a sample r ~ Unif[0,1] then g = a + b * log( -1 * log(1 - r)) follows g ~ Gumbel(a,b).

        uniform_samples = np.random.rand(self.num_samples)
        gumbel_samples = np.log(-1 * np.log(1 - uniform_samples)) * b + a
        self.mins = gumbel_samples

    def _required_parameters_initialized(self):
        """
        Checks if all required parameters are initialized.
        """
        return self.mins is not None

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Computes the information gain, i.e the predicted change in entropy of p_min (the distribution
        of the minimal value of the objective function) if we evaluate x.
        :param x: points where the acquisition is evaluated.
        """
        if not self._required_parameters_initialized():
            self.update_parameters()

        # Calculate GP posterior at candidate points
        fmean, fvar = self.model.predict(x)
        fsd = np.sqrt(fvar)
        # Clip below to improve numerical stability
        fsd = np.maximum(fsd, 1e-10)

        # standardise
        gamma = (self.mins - fmean) / fsd

        minus_cdf = 1 - norm.cdf(gamma)
        # Clip  to improve numerical stability
        minus_cdf = np.clip(minus_cdf, a_min=1e-10, a_max=1)

        # calculate monte-carlo estimate of information gain
        f_acqu_x = np.mean(-gamma * norm.pdf(gamma) / (2 * minus_cdf) - np.log(minus_cdf), axis=1)
        return f_acqu_x.reshape(-1, 1)

    @property
    def has_gradients(self) -> bool:
        """Returns that this acquisition has gradients"""
        return False


def _fit_gumbel(fmean, fsd):
    """
    Helper function to fit gumbel distribution when initialising the MES and MUMBO acquisition functions.

    The Gumbel distribution for minimas has a cumulative density function of f(y)= 1 - exp(-1 * exp((y - a) / b)), i.e. the q^th quantile is given by
    Q(q) = a + b * log( -1 * log(1 - q)). We choose values for a and b that match the Gumbel's
    interquartile range with that of the observed empirical cumulative density function of Pr(y*<y)
    i.e.  Pr(y* < lower_quantile)=0.25 and Pr(y* < upper_quantile)=0.75.
    """

    def probf(x: np.ndarray) -> float:
        # Build empirical CDF function
        return 1 - np.exp(np.sum(norm.logcdf(-(x - fmean) / fsd), axis=0))

    # initialise end-points for binary search (the choice of 5 standard deviations ensures that these are outside the IQ range)
    left = np.min(fmean - 5 * fsd)
    right = np.max(fmean + 5 * fsd)

    def binary_search(val: float) -> float:
        return bisect(lambda x: probf(x) - val, left, right, maxiter=10000)

    # Binary search for 3 percentiles
    lower_quantile, medium, upper_quantile = map(binary_search, [0.25, 0.5, 0.75])

    # solve for Gumbel scaling parameters
    b = (lower_quantile - upper_quantile) / (np.log(np.log(4.0 / 3.0)) - np.log(np.log(4.0)))
    a = medium - b * np.log(np.log(2.0))

    return a, b


class MUMBO(MaxValueEntropySearch):
    def __init__(
        self,
        model: Union[IModel, IEntropySearchModel],
        space: ParameterSpace,
        target_information_source_index: int = None,
        num_samples: int = 10,
        grid_size: int = 5000,
    ) -> None:

        """
        MUMBO acquisition function approximates the distribution of the value at the global
        minimum and tries to decrease its entropy.
        This is a multi-fidelity/multi-task extension of max-value entropy search (MES) suitiable
        for multi-information source problems where the objective function is the output of one
        of the information sources. The other information sources provide auxiliary
        information about the objective function
        See this paper for more details:
        Moss et al.
        MUMBO: MUlti-task Max-value Bayesian Optimisation
        ECML 2020

        :param model: GP model to compute the distribution of the minimum dubbed pmin.
        :param space: Domain space which we need for the sampling of the representer points
        :param target_information_source_index: The index of the information source we want to minimise
        :param num_samples: integer determining how many samples to draw of the minimum (does not need to be large)
        :param grid_size: number of random locations in grid used to fit the gumbel distribution and approximately generate
                          the samples of the minimum (recommend scaling with problem dimension, i.e. 10000*d)
        """

        if not isinstance(model, IEntropySearchModel):
            raise RuntimeError("Model is not supported for MES")

        # Find information source parameter in parameter space
        info_source_parameter, source_idx = _find_source_parameter(space)
        self.source_idx = source_idx

        # If not told otherwise assume we are in a multi-fidelity setting
        # and the highest index is the highest fidelity
        if target_information_source_index is None:
            target_information_source_index = max(info_source_parameter.domain)
        self.target_information_source_index = target_information_source_index

        self.model = model
        self.space = space
        self.num_samples = num_samples
        self.grid_size = grid_size

        # Initialize parameters to lazily compute them once needed
        self.mins = None

        super().__init__(model, space, num_samples, grid_size)

    def update_parameters(self):
        """
        MUMBO requires acces to a sample of possible minimum values y* of the objective function.
        To build this sample we approximate the empirical c.d.f of Pr(y*<y) with a Gumbel(a,b) distribution.
        This Gumbel distribution can then be easily sampled to yield approximate samples of y*

        This needs to be called once at the start of each BO step.
        """

        # First we generate a random grid of locations at which to fit the Gumbel distribution
        random_design = RandomDesign(self.space)
        grid = random_design.get_samples(self.grid_size)
        # also add the locations already queried in the previous BO steps
        grid = np.vstack([self.model.X, grid])
        # remove current fidelity index from sample
        grid = np.delete(grid, self.source_idx, axis=1)
        # Add objective function fidelity index to sample
        idx = np.ones((grid.shape[0])) * self.target_information_source_index
        grid = np.insert(grid, self.source_idx, idx, axis=1)
        # Get GP posterior at these points
        fmean, fvar = self.model.predict(grid)
        fsd = np.sqrt(fvar)

        # fit Gumbel distriubtion
        a, b = _fit_gumbel(fmean, fsd)

        # sample K times from this Gumbel distribution
        uniform_samples = np.random.rand(self.num_samples)
        gumbel_samples = -np.log(-np.log(uniform_samples)) * b + a
        self.mins = gumbel_samples

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Computes the information gain, i.e the change in entropy of p_min (the distribution
        of the minimal value of the objective function) if we would evaluate x.
        :param x: points where the acquisition is evaluated.
        """
        if not self._required_parameters_initialized():
            self.update_parameters()

        # Calculate GP posterior at candidate points
        fmean, fvar = self.model.predict(x)
        fsd = np.sqrt(fvar)
        # clip below for numerical stability
        fsd = np.maximum(fsd, 1e-10)

        # Also need posterior at locations with same search-space positions as x but on the objective function g
        # remove current fidelity index from sample
        x_target_fidelity = np.delete(x, self.source_idx, axis=1)
        # Add target fidelity index to sample
        idx = np.ones((x.shape[0])) * self.target_information_source_index
        x_target_fidelity = np.insert(x_target_fidelity, self.source_idx, idx, axis=1)
        gmean, gvar = self.model.predict(x_target_fidelity)
        gsd = np.sqrt(gvar)
        # clip below for numerical stability
        gsd = np.maximum(gsd, 1e-10)

        # also get pair-wise correlations between GP at x and x_target_fidelity
        # faster to do for loop rather than vectorize to avoid unecessary between term covariance calculations
        covariances = [
            self.model.get_covariance_between_points(x[i].reshape(1, -1), x_target_fidelity[i].reshape(1, -1))
            for i in range(0, x.shape[0])
        ]
        covariances = np.array(covariances).reshape(-1, 1)
        correlations = covariances / (fsd * gsd)
        # clip for numerical stability
        correlations = np.clip(correlations, -1, 1)

        # Calculate variance of extended skew Gaussian distributions (ESG)
        # These will be used to define reasonable ranges for the numerical
        # intergration of the ESG's differential entropy.
        gammas = (self.mins - fmean) / fsd
        minus_cdf = 1 - norm.cdf(gammas)
        # Clip  to improve numerical stability
        minus_cdf = np.clip(minus_cdf, a_min=1e-10, a_max=1)
        ESGmean = correlations * (norm.pdf(gammas)) / minus_cdf
        ESGvar = 1 + correlations * ESGmean * (gammas - norm.pdf(gammas) / minus_cdf)
        # Clip  to improve numerical stability
        ESGvar = np.maximum(ESGvar, 0)

        # get upper limits for numerical integration
        # we need this range to contain almost all of the ESG's probability density
        # we found +-8 standard deviations provides a tight enough approximation
        upper_limit = ESGmean + 8 * np.sqrt(ESGvar)
        lower_limit = ESGmean - 8 * np.sqrt(ESGvar)

        # perform numerical integrations
        # build discretisation
        z = np.linspace(lower_limit, upper_limit, num=5000)
        # calculate ESG density at these points
        minus_correlations = np.sqrt(1 - correlations**2)
        # clip below for numerical stability
        minus_correlations = np.maximum(minus_correlations, 1e-10)
        density = norm.pdf(z) * (1 - norm.cdf((gammas - correlations * z) / minus_correlations)) / minus_cdf
        # calculate point-wise entropy function contributions (carefuly where density is 0)
        entropy_function = -density * np.log(density, out=np.zeros_like(density), where=(density != 0))
        # perform integration over ranges
        approximate_entropy = simps(entropy_function.T, z.T)
        # build monte-carlo estimate over the gumbel samples
        approximate_entropy = np.mean(approximate_entropy, axis=0)

        # build MUMBO acquisition function
        f_acqu_x = 0.5 * np.log(2 * np.pi * np.e) - approximate_entropy
        return f_acqu_x.reshape(-1, 1)


def _find_source_parameter(space):
    # Find information source parameter in parameter space
    info_source_parameter = None
    source_idx = None
    for i, param in enumerate(space.parameters):
        if isinstance(param, InformationSourceParameter):
            info_source_parameter = param
            source_idx = i

    if info_source_parameter is None:
        raise ValueError("No information source parameter found in the parameter space")

    return info_source_parameter, source_idx



import numpy as np

from ...bayesian_optimization.acquisitions import (
    ExpectedImprovement,
    NegativeLowerConfidenceBound,
    ProbabilityOfImprovement,
)
from ...core import ParameterSpace
from ...core.loop import FixedIntervalUpdater, OuterLoop, SequentialPointCalculator
from ...core.loop.loop_state import create_loop_state
from ...core.optimization import AcquisitionOptimizerBase, GradientAcquisitionOptimizer
from ..models.bohamiann import Bohamiann
from ..models.random_forest import RandomForest
from .enums import AcquisitionType, ModelType


def create_bayesian_optimization_loop(
    x_init: np.ndarray,
    y_init: np.ndarray,
    parameter_space: ParameterSpace,
    acquisition_type: AcquisitionType,
    model_type: ModelType,
    cost_init: np.ndarray = None,
    model_kwargs: dict = None,
    acquisition_optimizer: AcquisitionOptimizerBase = None,
) -> OuterLoop:
    """
    Creates Bayesian optimization loop for Bayesian neural network or random forest models.

    :param x_init: 2d numpy array of shape (no. points x no. input features) of initial X data
    :param y_init: 2d numpy array of shape (no. points x no. targets) of initial Y data
    :param cost_init: 2d numpy array of shape (no. points x no. targets) of initial cost of each function evaluation
    :param parameter_space: parameter space describing input domain
    :param acquisition_type: an AcquisitionType enumeration object describing which acquisition function to use
    :param model_type: A ModelType enumeration object describing which model to use.
    :param model_kwargs: Key work arguments for the model constructor. See individual models for details.
    :param acquisition_optimizer: Optimizer selecting next evaluation points by maximizing acquisition.
                                  Gradient based optimizer is used if None. Defaults to None.
    :return: OuterLoop instance
    """

    if model_kwargs is None:
        model_kwargs = dict()

    # Create model
    if model_type is ModelType.RandomForest:
        model = RandomForest(x_init, y_init, **model_kwargs)
    elif model_type is ModelType.BayesianNeuralNetwork:
        model = Bohamiann(x_init, y_init, **model_kwargs)
    else:
        raise ValueError("Unrecognised model type: " + str(model_type))

    # Create acquisition
    if acquisition_type is AcquisitionType.EI:
        acquisition = ExpectedImprovement(model)
    elif acquisition_type is AcquisitionType.PI:
        acquisition = ProbabilityOfImprovement(model)
    elif acquisition_type is AcquisitionType.NLCB:
        acquisition = NegativeLowerConfidenceBound(model)
    else:
        raise ValueError("Unrecognised acquisition type: " + str(acquisition_type))

    if acquisition_optimizer is None:
        acquisition_optimizer = GradientAcquisitionOptimizer(parameter_space)
    candidate_point_calculator = SequentialPointCalculator(acquisition, acquisition_optimizer)

    if cost_init is None:
        loop_state = create_loop_state(x_init, y_init)
    else:
        loop_state = create_loop_state(x_init, y_init, cost=cost_init)

    model_updater = FixedIntervalUpdater(model, 1)
    return OuterLoop(candidate_point_calculator, model_updater, loop_state)



import logging
from typing import Callable, Dict, List, Tuple

import numpy as np
from numpy import linalg as la


def random_sample(bounds: np.ndarray, k: int) -> np.ndarray:
    """
    Generate a set of k n-dimensional points sampled uniformly at random
    :param bounds: n x 2 dimenional array containing upper/lower bounds for each dimension
    :param k: number of samples
    :return: k x n array containing the sampled points
    """

    # k: Number of points
    n = len(bounds)  # Dimensionality of each point
    X = np.zeros((k, n))
    for i in range(n):
        X[:, i] = np.random.uniform(bounds[i][0], bounds[i][1], k)

    return X


def static_sample(bounds: np.ndarray) -> np.ndarray:
    """
    Generate two locations outside the bounds

    :param bounds: n x 2 dimenional array containing upper/lower bounds for each dimension
    :return: 2 x n array containing the sampled points
    """
    n = len(bounds)  # Dimensionality of each point
    X = np.zeros((2, n))
    for i in range(n):
        X[0, i] = bounds[i][1] + 5 * (bounds[i][1] - bounds[i][0])
        X[1, i] = bounds[i][0] - 5 * (bounds[i][1] - bounds[i][0])
    return X


def grid_sample(n: int) -> np.ndarray:
    """
    Get a grid with equally distanced points

    :param n: dimensionality of the data
    :return: 2^n samples that are placed in a n dimendional grid
    """
    if n == 1:
        ret = np.array([[0], [1]])
    if n == 2:
        ret = np.array([[0, 0], [0, 1], [1, 0], [1, 1]])
    if n == 3:
        ret = np.array([[0, 0, 0], [0, 0, 1], [0, 1, 0], [0, 1, 1], [1, 0, 0], [1, 0, 1], [1, 1, 0], [1, 1, 1]])
    if n == 4:
        ret = np.array(
            [
                [0, 0, 0, 0],
                [0, 0, 0, 1],
                [0, 0, 1, 0],
                [0, 0, 1, 1],
                [0, 1, 0, 0],
                [0, 1, 0, 1],
                [0, 1, 1, 0],
                [0, 1, 1, 1],
                [1, 0, 0, 0],
                [1, 0, 0, 1],
                [1, 0, 1, 0],
                [1, 0, 1, 1],
                [1, 1, 0, 0],
                [1, 1, 0, 1],
                [1, 1, 1, 0],
                [1, 1, 1, 1],
            ]
        )
    if n == 5:
        ret = np.array(
            [
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 1],
                [0, 0, 0, 1, 0],
                [0, 0, 0, 1, 1],
                [0, 0, 1, 0, 0],
                [0, 0, 1, 0, 1],
                [0, 0, 1, 1, 0],
                [0, 0, 1, 1, 1],
                [0, 1, 0, 0, 0],
                [0, 1, 0, 0, 1],
                [0, 1, 0, 1, 0],
                [0, 1, 0, 1, 1],
                [0, 1, 1, 0, 0],
                [0, 1, 1, 0, 1],
                [0, 1, 1, 1, 0],
                [0, 1, 1, 1, 1],
                [1, 0, 0, 0, 0],
                [1, 0, 0, 0, 1],
                [1, 0, 0, 1, 0],
                [1, 0, 0, 1, 1],
                [1, 0, 1, 0, 0],
                [1, 0, 1, 0, 1],
                [1, 0, 1, 1, 0],
                [1, 0, 1, 1, 1],
                [1, 1, 0, 0, 0],
                [1, 1, 0, 0, 1],
                [1, 1, 0, 1, 0],
                [1, 1, 0, 1, 1],
                [1, 1, 1, 0, 0],
                [1, 1, 1, 0, 1],
                [1, 1, 1, 1, 0],
                [1, 1, 1, 1, 1],
            ]
        )
    return (ret) * (0.5) + 0.25


def give_comparisons(func: Callable, points: np.ndarray, si: int = 0) -> List[List[Tuple[int, int]]]:
    """
    Evaluates the function in given batch and returns the batch winner in a form accepted by the inference algorithms

    :param func: the black box objective
    :param points: batch locations
    :param si: starting index of the batch (the number of points evaluated so far)
    :return: Compar
    """
    ind = np.random.permutation(points.shape[0])
    y = func(points).reshape((-1, 1))
    s1, s2 = ind[::2], ind[1::2]
    return [[(s1[i] + si, s2[i] + si)] if y[s1[i]] < y[s2[i]] else [(s1[i] + si, s2[i] + si)] for i in range(len(s1))]


def boundit(x: np.ndarray, bounds: np.ndarray) -> np.ndarray:
    """
    Projects input to be within the bounds

    :param x: The point to be projected
    :param bounds: bounds of the hyper rectangle the point should be within
    :return: the point projected to be within the inputs
    """
    for i, bound in enumerate(bounds):
        if x[i] < bound[0]:
            x[i] = bound[0]
        elif x[i] > bound[1]:
            x[i] = bound[1]
    return x


def running_mean(x: np.ndarray, N: int) -> float:
    """
    Computes the running mean of N members for the array. The returned array is of length len(x)-N

    :param x: the input array for which we want to compute the running mean
    :param N: The length of the running mean window
    :return: The running mean of the array. Te returned array is of length len(x)-N
    """
    cumsum = np.cumsum(np.insert(x, 0, 0))
    return (cumsum[N:] - cumsum[:-N]) / float(N)


def max_difference(xs: np.ndarray):
    """
    Computes the maximum change in the input when iterating through it and moving from the previus value to the next one

    :param xs: array from which the max difference is wanted to be computed
    :return: maximum difference between the proceeding indices
    """
    min_elem = xs[0]
    max_elem = xs[0]
    max_diff = -1

    for elem in xs[1:]:
        min_elem = min(elem, min_elem)
        if elem > max_elem:
            max_elem = elem
    max_diff = max(max_diff, max_elem - min_elem)
    return max_diff


def adam(
    objective: Callable,
    x0: np.ndarray,
    args: Dict,
    max_it: int = 500,
    alpha: float = 0.01,
    beta1: float = 0.9,
    beta2: float = 0.999,
    eps: float = 1e-8,
    tol: float = 1e-3,
    lag: int = 100,
    bounds: np.ndarray = None,
    Nw: int = 20,
    get_logger: Callable = None,
) -> Tuple[float, float, float]:
    """
    Runs adam optimization algorithm for the objective

    :param objective: The function to be optimized
    :param x0: starting location
    :param args: additional arguments for objective
    :param max_it: maximum number of iterations
    :param alpha: learning rate
    :param beta1: tuning parameter between [0,1], generally close to 1
    :param beta2: tuning parameter between [0,1], generally close to 1
    :param eps: a pparameter that helps avoid division by zero when computing the new parameter
    :param tol: tolerance the optimization is ended after which the running mean goes
    :param lag: how many iterations are waited and the change in the running mean must be less than the tolerance
    :param bounds: optimization bounds
    :param Nw: how many iterations are used t ovompute the running mean
    :param get_logger: Function for receiving the legger where the prints are forwarded.
    :return: A tuple that contain the optimized location, mean of the 5 last iterations, and the whole optimization history of points
    """
    x = x0
    c = 0
    mt, vt = 0.0, 0.0  # moving average of gradient and squared gradient
    fs = []
    xs = [x0]
    if get_logger is not None:
        f, grad = objective(x, args)
        get_logger().debug("Starting from: {} with acq {}".format(x, f))
    it = 0
    while it < max_it:
        try:
            f, grad = objective(x, args)
        except:
            if get_logger is not None:
                get_logger().error("Error in objective when optimizing with Adam. Location:\n".format(x))
            break
        mt = beta1 * mt + (1 - beta1) * grad  # update the moving averages of the gradient
        vt = beta2 * vt + (1 - beta2) * (grad**2)  # updates the moving averages of the squared gradient
        m_cap = mt / (1 - (beta1 ** (it + 1)))  # calculates the bias-corrected estimates
        v_cap = vt / (1 - (beta2 ** (it + 1)))  # calculates the bias-corrected estimates
        delta = (alpha * m_cap) / (np.sqrt(v_cap) + eps)
        x = x - delta
        if bounds is not None:
            x = boundit(x, bounds)
        fs += [f]
        xs += [x]
        if (it > lag + Nw - 1) and (max_difference(running_mean(fs, Nw)[-lag:]) < tol):
            break
        it += 1
    if get_logger is not None:
        get_logger().debug("Ended to: {} with acq {} and gradient {} in {} iterations".format(x, f, grad, it))
    fsr = np.mean(fs[-5:]) if len(fs) > 4 else np.mean(fs)
    return x, fsr, xs  # return the mean of the 5 previous values to reduce stochasticity


def configure_logger(log_file=None) -> None:
    """
    Configures the file the prints are forwarded to

    :param log_file: The file the prints are forwarded to
    """
    if log_file is None:
        logger = logging.getLogger("logger.log")  # default logger
        file_handler = logging.StreamHandler()
    else:
        logger = logging.getLogger(log_file)
        formatter = logging.Formatter("%(asctime)s - %(levelname)-8s \n %(message)s")
        file_handler = logging.FileHandler(log_file, mode="w")
        file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)
    logger.setLevel(logging.DEBUG)
    logger.propagate = False


def comparison_form(y: np.ndarray) -> List[Tuple[int, int]]:
    """
    Transforms the already existing observations to the comparison form

    :param y: function evaluations in a batch
    :return: comparison form for the outputs
    """
    m = len(y)
    assert m > 1
    ind = np.argsort(y.flatten())
    return [(ind[0], ind[i]) for i in range(1, m)]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import List, Tuple

import numpy as np

from emukit.core.loop.user_function import UserFunctionWrapper


def hennig1D() -> Tuple[UserFunctionWrapper, List[Tuple[float, float]]]:
    r"""1D toy integrand coined by Philipp Hennig.

    One of the earlier mentions e.g., in `this talk <https://youtu.be/tZ9CP-kQAVI?t=704>`_
    (external link).

    .. math::
        f(x) = e^{-x^2 -\sin^2(3x)}

    :return: The wrapped test function, and the integrals bounds
             (the latter default to [-3, 3]).
    """
    integral_bounds = [(-3.0, 3.0)]
    return UserFunctionWrapper(_hennig1D), integral_bounds


def _hennig1D(x: np.ndarray) -> np.ndarray:
    """
    :param x: Locations for evaluation (num_points, 1).
    :return: The function values at x, shape (num_points, 1).
    """
    return np.exp(-(x**2) - np.sin(3.0 * x) ** 2)



"""
This file contains the multi-fidelity deep Gaussian process model from:
Deep Gaussian Processes for Multi-fidelity Modeling (Kurt Cutajar, Mark Pullin, Andreas Damianou, Neil Lawrence, Javier González)

The class intended for public consumption is MultiFidelityDeepGP, which is an emukit model class.

This file requires the following packages:
- tensorflow 1.x
- gpflow 1.x
- doubly_stochastic_dgp https://github.com/ICL-SML/Doubly-Stochastic-DGP/tree/master/doubly_stochastic_dgp
"""
import logging
from typing import List, Tuple

import numpy as np
from doubly_stochastic_dgp.utils import BroadcastingLikelihood
from gpflow import ParamList, autoflow, params_as_tensors, settings
from gpflow.actions import Action, Loop
from gpflow.kernels import RBF, Linear, White
from gpflow.likelihoods import Gaussian
from gpflow.mean_functions import Zero
from gpflow.models.model import Model
from gpflow.params import DataHolder, Minibatch
from gpflow.training import AdamOptimizer

from ...core.interfaces import IModel
from ...multi_fidelity.convert_lists_to_array import convert_x_list_to_array, convert_y_list_to_array

# Import packages that are not required by emukit and throw warning if they are not installed
try:
    import tensorflow as tf
except ImportError:
    raise ImportError(
        "tensorflow is not installed. Please installed version 1.8 by running pip install tensorflow==1.8"
    )

try:
    from doubly_stochastic_dgp.layers import SVGP_Layer
except ImportError:
    raise ImportError(
        "doubly_stochastic_dgp is not installed. "
        "Please run pip install git+https://github.com/ICL-SML/Doubly-Stochastic-DGP.git"
    )

try:
    import gpflow
except ImportError:
    raise ImportError("gpflow is not installed. Please run pip install gpflow==1.1.1")


float_type = settings.float_type
_log = logging.getLogger(__name__)


def init_layers_mf(Y, Z, kernels, num_outputs=None, Layer=SVGP_Layer):
    """
    Creates layer objects from initial data

    :param Y: Numpy array of training targets
    :param Z: List of numpy arrays of inducing point locations for each layer
    :param kernels: List of kernels for each layer
    :param num_outputs: Number of outputs (same for each layer)
    :param Layer: The layer object to use
    :return: List of layer objects with which to build a multi-fidelity deep Gaussian process model
    """
    num_outputs = num_outputs or Y[-1].shape[1]

    layers = []
    num_layers = len(Z)

    for i in range(0, num_layers):
        layers.append(Layer(kernels[i], Z[i], num_outputs, Zero()))
    return layers


class DGP_Base(Model):
    """
    The base class for Deep Gaussian process models.

    Implements a Monte-Carlo variational bound and convenience functions.
    """

    def __init__(self, X, Y, likelihood, layers, minibatch_size=None, num_samples=1, **kwargs):
        """

        :param X: List of training inputs where each element of the list is a numpy array corresponding to the inputs of one fidelity.
        :param Y: List of training targets where each element of the list is a numpy array corresponding to the inputs of one fidelity.
        :param likelihood: gpflow likelihood object for use at the final layer
        :param layers: List of doubly_stochastic_dgp.layers.Layer objects
        :param minibatch_size: Minibatch size if using minibatch trainingz
        :param num_samples: Number of samples when propagating predictions through layers
        :param kwargs: kwarg inputs to gpflow.models.Model
        """

        Model.__init__(self, **kwargs)

        self.Y_list = Y
        self.X_list = X
        self.minibatch_size = minibatch_size

        self.num_samples = num_samples

        # This allows a training regime where the first layer is trained first by itself, then the subsequent layer
        # and so on.
        self._train_upto_fidelity = -1

        if minibatch_size:
            for i, (x, y) in enumerate(zip(X, Y)):
                setattr(self, "num_data" + str(i), x.shape[0])
                setattr(self, "X" + str(i), Minibatch(x, minibatch_size, seed=0))
                setattr(self, "Y" + str(i), Minibatch(y, minibatch_size, seed=0))
        else:
            for i, (x, y) in enumerate(zip(X, Y)):
                setattr(self, "num_data" + str(i), x.shape[0])
                setattr(self, "X" + str(i), DataHolder(x))
                setattr(self, "Y" + str(i), DataHolder(y))

        self.num_layers = len(layers)
        self.layers = ParamList(layers)

        self.likelihood = BroadcastingLikelihood(likelihood)

    @params_as_tensors
    def propagate(self, X, full_cov=False, S=1, zs=None):
        """
        Propagate some prediction to the final layer and return predictions at each intermediate layer

        :param X: Input(s) at which to predict at
        :param full_cov: Whether the predict with the full covariance matrix
        :param S: Number of samples to use for sampling at intermediate layers
        :param zs: ??
        :return:
        """
        sX = tf.tile(tf.expand_dims(X, 0), [S, 1, 1])

        Fs, Fmeans, Fvars = [], [], []

        F = sX
        zs = zs or [
            None,
        ] * len(self.layers)

        for i, (layer, z) in enumerate(zip(self.layers, zs)):
            if i == 0:
                F, Fmean, Fvar = layer.sample_from_conditional(F, z=z, full_cov=full_cov)
            else:
                """

                KC - At all layers 1..L, the input to the next layer is original input augmented with
                the realisation of the function at the previous layer at that input.

                """
                F_aug = tf.concat([sX, F], 2)
                F, Fmean, Fvar = layer.sample_from_conditional(F_aug, z=z, full_cov=full_cov)

            Fs.append(F)
            Fmeans.append(Fmean)
            Fvars.append(Fvar)

        return Fs, Fmeans, Fvars

    @params_as_tensors
    def _build_predict(self, X, full_cov=False, S=1, fidelity=None):
        """
        Predicts from the fidelity level specified. If fidelity is not specified, return prediction at highest fidelity.

        :param X: Location at which to predict
        :param full_cov: Whether to predict full covariance matrix
        :param S: Number of samples to use for MC sampling between layers
        :param fidelity: zero based fidelity index at which to predict
        :return: (mean, variance) where each is [S, N, 1] where S is number of samples and N is number of predicted points.
        """

        if fidelity is None:
            fidelity = -1

        _, Fmeans, Fvars = self.propagate(X, full_cov=full_cov, S=S)
        return Fmeans[fidelity], Fvars[fidelity]

    def _likelihood_at_fidelity(self, Fmu, Fvar, Y, variance):
        """
        Calculate likelihood term for observations corresponding to one fidelity

        :param Fmu: Posterior mean
        :param Fvar: Posterior variance
        :param Y: training observations
        :param variance: likelihood variance
        :return:
        """
        return -0.5 * np.log(2 * np.pi) - 0.5 * tf.log(variance) - 0.5 * (tf.square(Y - Fmu) + Fvar) / variance

    def E_log_p_Y(self, X_f, Y_f, fidelity=None):
        """
        Calculate the expectation of the data log likelihood under the variational distribution with MC samples

        :param X_f: Training inputs for a given
        :param Y_f:
        :param fidelity:
        :return:
        """

        Fmean, Fvar = self._build_predict(X_f, full_cov=False, S=self.num_samples, fidelity=fidelity)

        if fidelity == (self.num_layers - 1):
            """
            KC - The likelihood of the observations at the last layer is computed using the model's 'likelihood' object
            """
            var_exp = self.likelihood.variational_expectations(Fmean, Fvar, Y_f)  # S, N, D
        else:
            """
            KC - The Gaussian likelihood of the observations at the intermediate layers is computed using the noise
            parameter pertaining to the White noise kernel.

            This assumes that a White kernel should be added to all layers except for the last!
            If no noise is desired, the variance parameter in the White kernel should be set to zero and fixed.
            """
            variance = self.layers[fidelity].kern.kernels[-1].variance

            f = lambda vars_SND, vars_ND, vars_N: self._likelihood_at_fidelity(
                vars_SND[0], vars_SND[1], vars_ND[0], vars_N
            )

            var_exp = f([Fmean, Fvar], [tf.expand_dims(Y_f, 0)], variance)

        return tf.reduce_mean(var_exp, 0)  # N, D

    @params_as_tensors
    def _build_likelihood(self):
        """
        ELBO calculation
        :return: MC estimate of lower bound
        """
        L = 0.0
        KL = 0.0
        for fidelity in range(self.num_layers):

            if (self._train_upto_fidelity != -1) and (fidelity > self._train_upto_fidelity):
                continue

            X_l = getattr(self, "X" + str(fidelity))
            Y_l = getattr(self, "Y" + str(fidelity))

            n_data = getattr(self, "num_data" + str(fidelity))
            scale = tf.cast(n_data, float_type) / tf.cast(tf.shape(X_l)[0], float_type)

            L += tf.reduce_sum(self.E_log_p_Y(X_l, Y_l, fidelity)) * scale
            KL += tf.reduce_sum(self.layers[fidelity].KL())

        self.L = L
        self.KL = KL

        return self.L - self.KL

    @autoflow((float_type, [None, None]), (tf.int32, []))
    def predict_f(self, Xnew, num_samples, fidelity=None):
        return self._build_predict(Xnew, full_cov=False, S=num_samples, fidelity=fidelity)

    @autoflow((float_type, [None, None]), (tf.int32, []))
    def predict_f_full_cov(self, Xnew, num_samples, fidelity=None):
        return self._build_predict(Xnew, full_cov=True, S=num_samples, fidelity=fidelity)

    @autoflow((float_type, [None, None]), (tf.int32, []))
    def predict_all_layers(self, Xnew, num_samples):
        return self.propagate(Xnew, full_cov=False, S=num_samples)

    @autoflow((float_type, [None, None]), (tf.int32, []))
    def predict_all_layers_full_cov(self, Xnew, num_samples):
        return self.propagate(Xnew, full_cov=True, S=num_samples)

    @autoflow((float_type, [None, None]), (tf.int32, []))
    def predict_y(self, Xnew, num_samples):
        Fmean, Fvar = self._build_predict(Xnew, full_cov=False, S=num_samples)
        return self.likelihood.predict_mean_and_var(Fmean, Fvar)

    @autoflow((float_type, [None, None]), (float_type, [None, None]), (tf.int32, []))
    def predict_density(self, Xnew, Ynew, num_samples):
        Fmean, Fvar = self._build_predict(Xnew, full_cov=False, S=num_samples)
        l = self.likelihood.predict_density(Fmean, Fvar, Ynew)
        log_num_samples = tf.log(tf.cast(num_samples, float_type))
        return tf.reduce_logsumexp(l - log_num_samples, axis=0)

    @classmethod
    def make_mf_dgp(cls, X, Y, Z, add_linear=True, minibatch_size=None):
        """
        Constructor for convenience. Constructs a mf-dgp model from training data and inducing point locations

        :param X: List of target
        :param Y:
        :param Z:
        :param add_linear:
        :return:
        """

        n_fidelities = len(X)

        Din = X[0].shape[1]
        Dout = Y[0].shape[1]

        kernels = [RBF(Din, active_dims=list(range(Din)), variance=1.0, lengthscales=1, ARD=True)]
        for l in range(1, n_fidelities):
            D = Din + Dout
            D_range = list(range(D))
            k_corr = RBF(Din, active_dims=D_range[:Din], lengthscales=1, variance=1.0, ARD=True)
            k_prev = RBF(Dout, active_dims=D_range[Din:], variance=1.0, lengthscales=1.0)
            k_in = RBF(Din, active_dims=D_range[:Din], variance=1.0, lengthscales=1, ARD=True)
            if add_linear:
                k_l = k_corr * (k_prev + Linear(Dout, active_dims=D_range[Din:], variance=1.0)) + k_in
            else:
                k_l = k_corr * k_prev + k_in
            kernels.append(k_l)

        """
        A White noise kernel is currently expected by Mf-DGP at all layers except the last.
        In cases where no noise is desired, this should be set to 0 and fixed, as follows:

            white = White(1, variance=0.)
            white.variance.trainable = False
            kernels[i] += white
        """
        for i, kernel in enumerate(kernels[:-1]):
            kernels[i] += White(1, variance=1e-6)

        num_data = 0
        for i in range(len(X)):
            _log.info("\nData at Fidelity {}".format(i + 1))
            _log.info("X - {}".format(X[i].shape))
            _log.info("Y - {}".format(Y[i].shape))
            _log.info("Z - {}".format(Z[i].shape))
            num_data += X[i].shape[0]

        layers = init_layers_mf(Y, Z, kernels, num_outputs=Dout)

        model = DGP_Base(X, Y, Gaussian(), layers, num_samples=10, minibatch_size=minibatch_size)

        return model

    def multi_step_training(self, n_iter=5000, n_iter_2=15000):
        """
        Train with variational covariance fixed to be small first, then free up and train covariance alongside other
        parameters. Inducing point locations are fixed throughout.
        """
        for layer in self.layers[:-1]:
            layer.q_sqrt = layer.q_sqrt.value * 1e-8
            layer.q_sqrt.trainable = False
        self.layers[-1].q_sqrt = self.layers[-1].q_sqrt.value * self.Y_list[-1].var() * 0.01
        self.layers[-1].q_sqrt.trainable = False
        self.likelihood.likelihood.variance = self.Y_list[-1].var() * 0.01
        self.likelihood.likelihood.variance.trainable = False

        # Run with covariance fixed
        self.run_adam(3e-3, n_iter)

        # Run with covariance free
        self.likelihood.likelihood.variance.trainable = True

        for layer in self.layers:
            layer.q_sqrt.trainable = True

        self.run_adam(1e-3, n_iter_2)

    def fix_inducing_point_locations(self):
        """
        Fix all inducing point locations
        """
        for layer in self.layers:
            layer.feature.Z.trainable = False

    def run_adam(self, lr, iterations):
        adam = AdamOptimizer(lr).make_optimize_action(self)
        actions = [adam, PrintAction(self, "MF-DGP with Adam")]
        loop = Loop(actions, stop=iterations)()
        self.anchor(self.enquire_session())


class PrintAction(Action):
    """
    For progress printing during optimization
    """

    def __init__(self, model, text):
        self.model = model
        self.text = text

    def run(self, ctx):
        if ctx.iteration % 2000 == 0:
            objective = ctx.session.run(self.model.objective)
            _log.info("ELBO {:.4f};  KL {:.4f}".format(ctx.session.run(self.model.L), ctx.session.run(self.model.KL)))
            _log.info("{}: iteration {} objective {:.4f}".format(self.text, ctx.iteration, objective))


class MultiFidelityDeepGP(IModel):
    """
    Inducing points are fixed for first part of optimization then freed.
    Both sets of inducing points are initialized at low fidelity training data locations.
    """

    def __init__(self, X, Y, Z=None, n_iter=5000, fix_inducing=True, multi_step_training=True, minibatch_size=None):
        self._Y = Y
        self._X = X
        self.minibatch_size = minibatch_size

        if Z is None:
            self.Z = self._make_inducing_points(X, Y)
        else:
            self.Z = Z

        self.model = self._get_model(X, Y, self.Z)
        self.name = "mfdgp"
        self.n_fidelities = len(X)
        self.n_iter = n_iter
        self.fix_inducing = fix_inducing
        self.model.fix_inducing_point_locations()
        self.multi_step_training = multi_step_training

    def set_data(self, X: np.ndarray, Y: np.ndarray) -> None:
        raise NotImplementedError()

    def _get_model(self, X, Y, Z):
        return DGP_Base.make_mf_dgp(X, Y, Z, minibatch_size=self.minibatch_size)

    def predict(self, X: np.array) -> Tuple[np.array, np.array]:
        # assume high fidelity only!!!!
        assert np.all(X[:, -1] == (self.n_fidelities - 1))

        x_test = X[:, :-1]
        y_m, y_v = self.model.predict_y(x_test, 250)
        y_mean_high = np.mean(y_m, axis=0).flatten()
        y_var_high = np.mean(y_v, axis=0).flatten() + np.var(y_m, axis=0).flatten()
        return y_mean_high[:, None], y_var_high[:, None]

    def optimize(self) -> None:
        """
        Optimize variational parameters alongside kernel and likelihood parameters using the following regime:
            1. Optimize the parameters while fixing the intermediate layer mean variational parameters
            2. Free the mean of the variational distribution and optimize all parameters together
        """

        if self.multi_step_training:
            _log.info("\n--- Optimization: {} ---\n".format(self.name))
            self.model.layers[0].q_mu = self._Y[0]
            for i, layer in enumerate(self.model.layers[1:-1]):
                layer.q_mu = self._Y[i][::2]
                layer.q_mu.trainable = False

            self.model.fix_inducing_point_locations()
            self.model.multi_step_training(self.n_iter)

            for layer in self.model.layers[:-1]:
                layer.q_mu.trainable = True
            self.model.run_adam(1e-3, 15000)
        else:
            self.model.run_adam(1e-3, 20000)

    @staticmethod
    def _make_inducing_points(X: List, Y: List) -> List:
        """
        Makes inducing points at every other training point location which is deafult behaviour if no inducing point
        locations are passed

        :param X: training locations
        :param Y: training targets
        :return: List of numpy arrays containing inducing point locations
        """
        Z = [X[0].copy()]
        for x, y in zip(X[:-1], Y[:-1]):
            Z.append(np.concatenate((x.copy()[::2], y.copy()[::2]), axis=1))
        return Z

    @property
    def X(self) -> np.array:
        return convert_x_list_to_array(self._X)

    @property
    def Y(self) -> np.array:
        return convert_y_list_to_array(self._Y)



import logging
from typing import Callable, Optional

import numpy as np

_log = logging.getLogger(__name__)


class IConstraint:
    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        :param x: Array of shape (n_points x n_dims) containing input locations to evaluate constraint at
        :return: Numpy array of shape (n_input,) where an element will be 1 if the corresponding input satisfies the
                 constraint and zero if the constraint is violated
        """
        raise NotImplementedError


class InequalityConstraint(IConstraint):
    def __init__(self, lower_bound: np.ndarray, upper_bound: np.ndarray):
        """
        :param lower_bound: Lower bound vector of size (n_constraint,). Can be -np.inf for one sided constraint
        :param upper_bound: Upper bound vector of size (n_constraint,). Can be np.inf for one sided constraint
        """
        if (lower_bound is None) and (upper_bound is None):
            raise ValueError("Neither lower nor upper bounds is set, at least one must be specified")

        # Default lower bound to -infinity
        if lower_bound is None:
            lower_bound = np.full([upper_bound.shape[0]], -np.inf)

        # Default upper bound to +infinity
        if upper_bound is None:
            upper_bound = np.full([lower_bound.shape[0]], np.inf)

        if np.any((lower_bound == -np.inf) & (upper_bound == np.inf)):
            raise ValueError("One or more inequality constraints are unbounded")

        if np.any(lower_bound >= upper_bound):
            raise ValueError("Lower bound is greater than or equal to upper bound for one or more constraints")

        self.lower_bound = lower_bound
        self.upper_bound = upper_bound


class LinearInequalityConstraint(InequalityConstraint):
    """
    Constraint of the form lower_bound <= Ax <= upper_bound where the matrix A is called "constraint_matrix"
    """

    def __init__(self, constraint_matrix: np.ndarray, lower_bound: np.ndarray = None, upper_bound: np.ndarray = None):
        """

        :param constraint_matrix: (n_constraint, n_x_dims) matrix in b_lower <= Ax <= b_upper
        :param lower_bound: Lower bound vector of size (n_constraint,). Can be -np.inf for one sided constraint
        :param upper_bound: Upper bound vector of size (n_constraint,). Can be np.inf for one sided constraint
        """
        super().__init__(lower_bound, upper_bound)
        if (constraint_matrix.shape[0] != lower_bound.shape[0]) or (constraint_matrix.shape[0] != upper_bound.shape[0]):
            raise ValueError(
                "Shape mismatch between constraint matrix {} and lower {} or upper {} bounds".format(
                    constraint_matrix.shape, lower_bound.shape, upper_bound.shape
                )
            )

        self.constraint_matrix = constraint_matrix

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluate whether constraints are violated or satisfied at a set of x locations

        :param x: Array of shape (n_points x n_dims) containing input locations to evaluate constraint at
        :return: Numpy array of shape (n_points, ) where an element will be 1 if the corresponding input satisfies all
                 constraints and zero if any constraint is violated
        """
        if self.constraint_matrix.shape[1] != x.shape[1]:
            raise ValueError(
                "Dimension mismatch between constraint matrix (second dim {})"
                + " and input x (second dim {})".format(self.constraint_matrix.shape[1], x.shape[1])
            )

        # Transpose here is needed to handle input dimensions
        # that is, A is (n_const, n_dims) and x is (n_points, n_dims)
        ax = self.constraint_matrix.dot(x.T).T
        return np.all((ax >= self.lower_bound) & (ax <= self.upper_bound), axis=1)


class NonlinearInequalityConstraint(InequalityConstraint):
    """
    Constraint of the form lower_bound <= g(x) <= upper_bound
    """

    def __init__(
        self,
        constraint_function: Callable,
        lower_bound: np.ndarray,
        upper_bound: np.ndarray,
        jacobian_fun: Optional[Callable] = None,
    ):
        """
        :param constraint_function: function defining constraint in b_lower <= fun(x) <= b_upper.
                                    Has signature f(x) -> array, shape(m,) where x is 1d and m is the number of constraints
        :param lower_bound: Lower bound vector of size (n_constraint,). Can be -np.inf for one sided constraint
        :param upper_bound: Upper bound vector of size (n_constraint,). Can be np.inf for one sided constraint
        :param jacobian_fun: Function returning the jacobian of the constraint function. Optional, if not supplied
                             the optimizer will use finite differences to calculate the gradients of the constraint
        """

        super().__init__(lower_bound, upper_bound)

        self.fun = constraint_function
        self.jacobian_fun = jacobian_fun

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluate whether constraints are violated or satisfied at a set of x locations

        :param x: Array of shape (n_points x n_dims) containing input locations to evaluate constraint at
        :return: Numpy array of shape (n_input,) where an element will be 1 if the corresponding input satisfies the
                 constraint and zero if the constraint is violated
        """
        fun_x = np.array([self.fun(x) for x in x])
        return np.all([fun_x >= self.lower_bound, fun_x <= self.upper_bound], axis=0)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .models import IDifferentiable  # noqa: F401
from .models import IJointlyDifferentiable  # noqa: F401
from .models import IModel  # noqa: F401
from .models import IModelWithNoise  # noqa: F401
from .models import IPriorHyperparameters  # noqa: F401



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import List, Tuple

import numpy as np

from emukit.core.loop.user_function import UserFunctionWrapper


def sombrero2D(freq: float = 1.0) -> Tuple[UserFunctionWrapper, List[Tuple[float, float]]]:
    r"""2D Sombrero function.

    .. math::
        f(x) = \frac{\operatorname{sin}(\pi r \omega)}{\pi r \omega}

    where :math:`\omega` is the :attr:`freq` parameter and :math:`r=\|x\|` is the length
    of the input vector :math:`x`.

    :param freq: The frequency of the sombrero (must be > 0, defaults to 1).
    :return: The wrapped test function, and the integrals bounds (the latter defaults to [-3, 3]^2).
    """
    func = lambda x: _sombrero2D(x, freq)
    integral_bounds = 2 * [(-3.0, 3.0)]
    return UserFunctionWrapper(func), integral_bounds


def _sombrero2D(x: np.ndarray, freq: float) -> np.ndarray:
    """
    :param x: Locations for evaluation (num_points, 2).
    :param freq: The frequency of the sombrero (must be > 0).
    :return: The function values at x, shape (num_points, 1).
    """
    r = np.sqrt((x * x).sum(axis=1))
    r_scaled = (np.pi * freq) * r
    result = np.sin(r_scaled) / r_scaled
    result[np.isnan(result)] = 1.0
    return result[:, np.newaxis]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .branin import branin_function  # noqa: F401
from .forrester import forrester_function, multi_fidelity_forrester_function  # noqa: F401
from .non_linear_sin import multi_fidelity_non_linear_sin  # noqa: F401
from .sixhumpcamel import sixhumpcamel_function  # noqa: F401



from copy import deepcopy
from typing import Tuple

import GPy
import numpy as np

from emukit.model_wrappers.gpy_model_wrappers import GPyModelWrapper


class FabolasKernel(GPy.kern.Kern):
    def __init__(self, input_dim, basis_func, a=1.0, b=1.0, active_dims=None):

        super(FabolasKernel, self).__init__(input_dim, active_dims, "fabolas_kernel")

        assert input_dim == 1

        self.basis_func = basis_func

        self.a = GPy.core.parameterization.Param("a", a)
        self.b = GPy.core.parameterization.Param("b", b)

        self.link_parameters(self.a, self.b)

    def K(self, X, X2):
        if X2 is None:
            X2 = X

        X_ = self.basis_func(X)
        X2_ = self.basis_func(X2)
        k = np.dot(X_ * self.b, X2_.T) + self.a

        return k

    def update_gradients_full(self, dL_dK, X, X2):
        if X2 is None:
            X2 = X
        X_ = self.basis_func(X)
        X2_ = self.basis_func(X2)
        self.a.gradient = np.sum(dL_dK)
        self.b.gradient = np.sum(np.dot(np.dot(X_, X2_.T), dL_dK))

    def Kdiag(self, X):
        return np.diag(self.K(X, X))


def linear(s):
    return s


def quad(s):
    return (1 - s) ** 2


def transform(s, s_min, s_max):
    s_transform = (np.log2(s) - np.log2(s_min)) / (np.log2(s_max) - np.log2(s_min))
    return s_transform


def retransform(s_transform, s_min, s_max):
    s = np.rint(2 ** (s_transform * (np.log2(s_max) - np.log2(s_min)) + np.log2(s_min)))
    return s


class FabolasModel(GPyModelWrapper):
    def __init__(
        self, X_init: np.ndarray, Y_init: np.ndarray, s_min: float, s_max: float, basis_func=linear, noise: float = 1e-6
    ):
        """
        Fabolas Gaussian processes model which models the validation error / cost of
        hyperparameter configurations across training dataset subsets.

        :param X_init: training data points
        :param Y_init: training targets
        :param basis_func: basis function which describes the change in performance across dataset subsets
        :param noise: observation noise added to the diagonal of the kernel matrix
        """

        self.noise = noise
        self.s_min = s_min
        self.s_max = s_max
        self._X = deepcopy(X_init)
        self._X[:, -1] = transform(self._X[:, -1], self.s_min, self.s_max)
        self._Y = Y_init
        self.basis_func = basis_func
        kernel = GPy.kern.Matern52(
            input_dim=self._X.shape[1] - 1,
            active_dims=[i for i in range(self._X.shape[1] - 1)],
            variance=np.var(self._Y),
            ARD=True,
        )
        kernel *= FabolasKernel(input_dim=1, active_dims=[self._X.shape[1] - 1], basis_func=basis_func)
        kernel += GPy.kern.White(input_dim=1, active_dims=[self._X.shape[1] - 1], variance=1e-6)

        gp = GPy.models.GPRegression(self._X, self._Y, kernel=kernel, noise_var=noise)
        gp.kern.set_prior(GPy.priors.Uniform(0, 5))
        gp.likelihood.constrain_positive()
        super(FabolasModel, self).__init__(gpy_model=gp, n_restarts=3)

    def predict(self, X):
        """
        :param X: (n_points x n_dimensions) array containing locations at which to get predictions
        :return: (mean, variance) Arrays of size n_points x 1 of the predictive distribution at each input location
        """
        X_ = deepcopy(X)
        X_[:, -1] = transform(X_[:, -1], self.s_min, self.s_max)
        return super(FabolasModel, self).predict(X_)

    def set_data(self, X, Y):
        """
        Sets training data in model

        :param X: New training features
        :param Y: New training outputs
        """
        self._X = deepcopy(X)
        self._X[:, -1] = transform(self._X[:, -1], self.s_min, self.s_max)
        self._Y = Y
        try:
            self.model.set_XY(self._X, self.Y)
        except:
            kernel = GPy.kern.Matern52(
                input_dim=self._X.shape[1] - 1,
                active_dims=[i for i in range(self._X.shape[1] - 1)],
                variance=np.var(self.Y),
                ARD=True,
            )
            kernel *= FabolasKernel(input_dim=1, active_dims=[self._X.shape[1] - 1], basis_func=self.basis_func)
            kernel *= GPy.kern.OU(input_dim=1, active_dims=[self._X.shape[1] - 1])

            self.model = GPy.models.GPRegression(self._X, self.Y, kernel=kernel, noise_var=self.noise)
            self.model.likelihood.constrain_positive()

    def get_f_minimum(self):
        """
        Predicts for all observed data points the validation error on the full dataset and returns
        the smallest mean prediciton

        :return: Array of size 1 x 1
        """
        proj_X = deepcopy(self._X)
        proj_X[:, -1] = np.ones(proj_X.shape[0]) * self.s_max
        mean_highest_dataset = self.model.predict(proj_X)

        return np.min(mean_highest_dataset, axis=0)

    @property
    def X(self):
        X = deepcopy(self._X)
        X[:, -1] = retransform(X[:, -1], self.s_min, self.s_max)
        return X

    @property
    def Y(self):
        return self._Y

    def get_prediction_gradients(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        :param X: (n_points x n_dimensions) array containing locations at which to get gradient of the predictions
        :return: (mean gradient, variance gradient) n_points x n_dimensions arrays of the gradients of the predictive
                 distribution at each input location
        """
        X_ = deepcopy(X)
        X_[:, -1] = transform(X_[:, -1], self.s_min, self.s_max)

        return super(FabolasModel, self).get_prediction_gradients(X_)

    def predict_covariance(self, X: np.ndarray, with_noise: bool = True) -> np.ndarray:
        """
        Calculates posterior covariance between points in X
        :param X: Array of size n_points x n_dimensions containing input locations to compute posterior covariance at
        :param with_noise: Whether to include likelihood noise in the covariance matrix
        :return: Posterior covariance matrix of size n_points x n_points
        """
        X_ = deepcopy(X)
        X_[:, -1] = transform(X_[:, -1], self.s_min, self.s_max)

        return super(FabolasModel, self).predict_covariance(X_, with_noise)

    def get_covariance_between_points(self, X1: np.ndarray, X2: np.ndarray) -> np.ndarray:
        """
        Calculate posterior covariance between two points
        :param X1: An array of shape 1 x n_dimensions that contains a data single point. It is the first argument of the
                   posterior covariance function
        :param X2: An array of shape n_points x n_dimensions that may contain multiple data points. This is the second
                   argument to the posterior covariance function.
        :return: An array of shape n_points x 1 of posterior covariances between X1 and X2
        """
        X_1 = deepcopy(X1)
        X_1[:, -1] = transform(X_1[:, -1], self.s_min, self.s_max)
        X_2 = deepcopy(X2)
        X_2[:, -1] = transform(X_2[:, -1], self.s_min, self.s_max)

        return super(FabolasModel, self).get_covariance_between_points(X1, X2)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np


class ICalculateVarianceReduction:
    def calculate_variance_reduction(self, x_train_new: np.ndarray, x_test: np.ndarray) -> np.ndarray:
        """
        Computes the variance reduction at x_test, if a new point at x_train_new is acquired
        """
        raise NotImplementedError



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np

from emukit.core import ContinuousParameter, InformationSourceParameter, ParameterSpace
from emukit.core.loop.user_function import MultiSourceFunctionWrapper


def multi_fidelity_forrester_function(high_fidelity_noise_std_deviation=0, low_fidelity_noise_std_deviation=0):
    """
    Two-level multi-fidelity forrester function where the high fidelity is given by:

    .. math::
        f(x) = (6x - 2)^2 \sin(12x - 4)

    and the low fidelity approximation given by:

    .. math::
        f_{low}(x) = 0.5 f_{high}(x) + 10 (x - 0.5) + 5

    :param high_fidelity_noise_std_deviation: Standard deviation of observation noise on high fidelity observations.
                                              Defaults to zero.
    :param low_fidelity_noise_std_deviation: Standard deviation of observation noise on low fidelity observations.
                                             Defaults to zero.
    :return: Tuple of user function object and parameter space object
    """
    parameter_space = ParameterSpace([ContinuousParameter("x", 0, 1), InformationSourceParameter(2)])
    user_function = MultiSourceFunctionWrapper(
        [
            lambda x: forrester_low(x, low_fidelity_noise_std_deviation),
            lambda x: forrester(x, high_fidelity_noise_std_deviation),
        ]
    )
    return user_function, parameter_space


def forrester_function(noise_standard_deviation=0):
    """
    Forrester function

    .. math::
        f(x) = (6x - 2)^2 \sin(12x - 4)

    :param noise_standard_deviation: Standard deviation of normally distributed observation noise
    :return: Tuple of function and parameter space object
    """

    def forrester_fcn(x):
        return forrester(x, sd=noise_standard_deviation)

    return forrester_fcn, ParameterSpace([ContinuousParameter("x", 0, 1)])


def forrester(x, sd=0):
    """
    Forrester function

    :param x: input vector to be evaluated
    :param sd: standard deviation of noise parameter
    :return: outputs of the function
    """
    x = x.reshape((len(x), 1))
    n = x.shape[0]
    fval = ((6 * x - 2) ** 2) * np.sin(12 * x - 4)
    if sd == 0:
        noise = np.zeros(n).reshape(n, 1)
    else:
        noise = np.random.normal(0, sd, n).reshape(n, 1)
    return fval.reshape(n, 1) + noise


def forrester_low(x, sd=0):
    """
    Low fidelity forrester function approximation:

    :param x: input vector to be evaluated
    :param sd: standard deviation of observation noise at low fidelity
    :return: outputs of the function
    """
    high_fidelity = forrester(x, 0)
    return 0.5 * high_fidelity + 10 * (x[:, [0]] - 0.5) + 5 + np.random.randn(x.shape[0], 1) * sd



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


"""
This file contains the "UserFunction" base class and implementations

The user function is the objective function in optimization, the integrand in quadrature or the function to be learnt
in experimental design.
"""

import abc
import logging
from typing import Callable, List

import numpy as np

from .user_function_result import UserFunctionResult

_log = logging.getLogger(__name__)


class UserFunction(abc.ABC, Callable):
    """The user supplied function is interrogated as part of the outer loop"""

    @abc.abstractmethod
    def evaluate(self, X: np.ndarray) -> List[UserFunctionResult]:
        pass

    def __call__(self, X: np.ndarray) -> List[UserFunctionResult]:
        return self.evaluate(X)


class UserFunctionWrapper(UserFunction):
    """Wraps a user-provided python function."""

    def __init__(self, f: Callable, extra_output_names: List[str] = None):
        """
        :param f: A python function that takes in a 2d numpy ndarray of inputs and returns a either a 2d numpy array
                  of function outputs or a tuple of (outputs, auxillary_output_1, auxilary_output_2, ...)
                  where all outputs are 2d
        :param extra_output_names: If the function f returns a tuple, the first output should be the value of the
                                   objective, which will be named "Y", names for subsequent outputs should be included
                                   in this list.
        """
        self.f = f
        self.extra_output_names = [] if extra_output_names is None else extra_output_names

    def evaluate(self, inputs: np.ndarray) -> List[UserFunctionResult]:
        """
        Evaluates python function by providing it with numpy types and converts the output
        to a List of UserFunctionResults

        :param inputs: List of function inputs at which to evaluate function
        :return: List of function results
        """
        if inputs.ndim != 2:
            raise ValueError(
                "User function should receive 2d array as an input, "
                "actual input dimensionality is {}".format(inputs.ndim)
            )

        _log.info("Evaluating user function for {} point(s)".format(inputs.shape[0]))
        outputs = self.f(inputs)

        if isinstance(outputs, tuple):
            user_fcn_outputs = outputs[0]
            extra_outputs = outputs[1:]
        elif isinstance(outputs, np.ndarray):
            user_fcn_outputs = outputs
            extra_outputs = tuple()
        else:
            raise ValueError(
                "User provided function should return a tuple or an ndarray, " "{} received".format(type(outputs))
            )

        # Validate number of outputs returned by the user function
        if len(extra_outputs) != len(self.extra_output_names):
            raise ValueError(
                "User function provided {} outputs but UserFunctionWrapper expected {}".format(
                    len(extra_outputs) + 1, len(self.extra_output_names) + 1
                )
            )

        if user_fcn_outputs.ndim != 2:
            raise ValueError(
                "User function should return 2d array or a tuple of 2d arrays as an output, "
                "actual output dimensionality is {}".format(outputs.ndim)
            )

        results = []
        for i in range(user_fcn_outputs.shape[0]):
            kwargs = dict([(name, val[i]) for name, val in zip(self.extra_output_names, extra_outputs)])
            results.append(UserFunctionResult(inputs[i], user_fcn_outputs[i], **kwargs))

        return results


class MultiSourceFunctionWrapper(UserFunction):
    """
    Wraps a list of python functions that each correspond to different information source.
    """

    def __init__(self, f: List, source_index: int = -1, extra_output_names: List[str] = None) -> None:
        """
        :param f: A list of python function that take in a 2d numpy ndarrays of inputs and return 2d numpy ndarrays
                  of outputs.
        :param source_index: An integer indicating which column of X contains the index of the information source.
                             Default to the last dimension of the input.
        :param extra_output_names: If the function f returns a tuple, the first output should be the value of the
                                   objective, which will be named "Y", names for subsequent outputs should be included
                                   in this list.
        """
        self.f = f
        self.source_index = source_index
        self.extra_output_names = [] if extra_output_names is None else extra_output_names

    def evaluate(self, inputs: np.ndarray) -> List[UserFunctionResult]:
        """
        Evaluates the python functions corresponding to the appropriate information source

        :param inputs: A list of inputs to evaluate the function at
                       with information source index appended as last column
        :return: A list of function outputs
        """

        if inputs.ndim != 2:
            raise ValueError(
                "User function should receive 2d array as an input, "
                "actual input dimensionality is {}".format(inputs.ndim)
            )

        n_sources = len(self.f)

        _log.info("Evaluating multi-source user function for {} point(s)".format(inputs.shape[0]))
        # Run each source function for all inputs at that source
        indices, outputs, extra_outputs = [], [], []
        source_indices = inputs[:, self.source_index]
        source_inputs = np.delete(inputs, self.source_index, axis=1)
        for i_source in range(n_sources):
            # Find inputs at that source
            this_source_input_indices = np.flatnonzero(source_indices == i_source)
            indices.append(this_source_input_indices)
            this_source_inputs = source_inputs[this_source_input_indices]
            this_outputs = self.f[i_source](this_source_inputs)

            if isinstance(this_outputs, tuple):
                outputs.append(this_outputs[0])
                extra_outputs.append(this_outputs[1:])

                # Check correct number of outputs from user function
                if len(extra_outputs[-1]) != len(self.extra_output_names):
                    raise ValueError(
                        "Expected {} outputs from user function but got {}".format(
                            len(self.extra_output_names) + 1, len(extra_outputs[-1]) + 1
                        )
                    )
            elif isinstance(this_outputs, np.ndarray):
                outputs.append(this_outputs)

                # Check correct number of outputs from user function
                if len(self.extra_output_names) != 0:
                    raise ValueError(
                        "Expected {} output from user function but got 1".format(len(self.extra_output_names) + 1)
                    )

                # Dummy extra outputs - won't be used below
                extra_outputs.append(tuple())
            else:
                raise ValueError(
                    "User provided function should return a tuple or an ndarray, "
                    "{} received".format(type(this_outputs))
                )

        sort_indices = np.argsort(np.concatenate(indices, axis=0))
        outputs = np.concatenate(outputs, axis=0)

        # Concatenate list of lists to single list
        n_extra_outputs = len(self.extra_output_names)
        extra_output_lists = [[] for _ in range(n_extra_outputs)]
        for i_source in range(n_sources):
            for i_output in range(n_extra_outputs):
                extra_output_lists[i_output].extend(extra_outputs[i_source][i_output])

        results = []
        for i, idx_sorted in enumerate(sort_indices):
            # Put extra outputs into a dictionary so we can pass them as key word arguments
            kwargs = dict([(name, val[idx_sorted]) for name, val in zip(self.extra_output_names, extra_output_lists)])
            results.append(UserFunctionResult(inputs[i], outputs[idx_sorted], **kwargs))
        return results



from typing import Tuple

import numpy as np

try:
    from sklearn.gaussian_process import GaussianProcessRegressor
except ImportError:
    ImportError("scikit-learn needs to be installed in order to use SklearnGPRWrapper")

from emukit.core.interfaces.models import IModel


class SklearnGPRWrapper(IModel):
    def __init__(self, sklearn_model: GaussianProcessRegressor):
        """
        :param sklearn_model: Scikit-learn GPR model to wrap
        """
        self.model = sklearn_model

    def predict(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Predict mean and variance values for given points

        :param X: array of shape (n_points x n_inputs) of points to run prediction for
        :return: Tuple of mean and variance which are 2d arrays of shape (n_points x n_outputs)
        """
        mean, std = self.model.predict(X, return_std=True)
        if mean.ndim == 1:
            mean = mean[:, None]
        return mean, np.power(std, 2.0).reshape(-1, 1)

    def set_data(self, X: np.ndarray, Y: np.ndarray) -> None:
        """
        Sets training data in model
        :param X: new points
        :param Y: function values at new points X

        """
        self.model.X_train_, self.model.y_train_ = X, Y

    def optimize(self) -> None:
        """
        Optimize hyper-parameters of model
        """
        self.model.fit(self.X, self.Y)

    @property
    def X(self) -> np.ndarray:
        return self.model.X_train_

    @property
    def Y(self) -> np.ndarray:
        return self.model.y_train_



"""Kernel embeddings for Bayesian quadrature."""

# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .quadrature_kernels import (  # isort: skip
    GaussianEmbedding,
    LebesgueEmbedding,
    QuadratureKernel,
    QuadratureProductKernel,
)
from .quadrature_brownian import (
    QuadratureBrownian,
    QuadratureBrownianLebesgueMeasure,
    QuadratureProductBrownian,
    QuadratureProductBrownianLebesgueMeasure,
)
from .quadrature_matern12 import QuadratureProductMatern12, QuadratureProductMatern12LebesgueMeasure
from .quadrature_matern32 import QuadratureProductMatern32, QuadratureProductMatern32LebesgueMeasure
from .quadrature_matern52 import QuadratureProductMatern52, QuadratureProductMatern52LebesgueMeasure
from .quadrature_rbf import QuadratureRBF, QuadratureRBFGaussianMeasure, QuadratureRBFLebesgueMeasure

__all__ = [
    "QuadratureKernel",
    "QuadratureProductKernel",
    "LebesgueEmbedding",
    "GaussianEmbedding",
    "QuadratureBrownian",
    "QuadratureBrownianLebesgueMeasure",
    "QuadratureRBF",
    "QuadratureRBFLebesgueMeasure",
    "QuadratureRBFGaussianMeasure",
    "QuadratureProductMatern52",
    "QuadratureProductMatern52LebesgueMeasure",
    "QuadratureProductMatern32",
    "QuadratureProductMatern32LebesgueMeasure",
    "QuadratureProductMatern12",
    "QuadratureProductMatern12LebesgueMeasure",
    "QuadratureProductBrownian",
    "QuadratureProductBrownianLebesgueMeasure",
]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import List

import numpy as np

from ...core.continuous_parameter import ContinuousParameter
from ..typing import BoundsType


class BoxDomain:
    r"""A box domain defined by a hyper-cube.

    :param bounds: The bounds defining the box.
                   List of :math:`d` tuples :math:`[(a_1, b_1), (a_2, b_2), \dots, (a_d, b_d)]`,
                   where :math:`d` is the dimensionality of the domain and the tuple :math:`(a_i, b_i)`
                   contains the lower and upper bound of dimension :math:`i` defining the box domain.
    :param name: Name of parameter.

    """

    def __init__(self, bounds: BoundsType, name: str = ""):
        self._check_bound_validity(bounds)
        self._bounds = bounds
        self.dim = len(bounds)
        self.name = name

        # set upper and lower bounds arrays for convenience
        bounds = np.array(bounds)
        self.lower_bounds = bounds[:, 0]
        self.upper_bounds = bounds[:, 1]

    @property
    def bounds(self) -> BoundsType:
        """The bounds defining the hypercube."""
        return self._bounds

    @bounds.setter
    def bounds(self, new_bounds: BoundsType) -> None:
        """Sets new box bounds and checks their validity.

        :param new_bounds: The new bounds.

        :raises ValueError: If ``new_bounds`` is not equal to dimensionality of measure.

        """
        if not len(new_bounds) == self.dim:
            raise ValueError("Length of new box bounds is {} (length {} expected).".format(len(new_bounds), self.dim))

        self._check_bound_validity(new_bounds)
        self._bounds = new_bounds

        new_bounds = np.array(new_bounds)
        self.lower_bounds = new_bounds[:, 0]
        self.upper_bounds = new_bounds[:, 1]

    def _check_bound_validity(self, bounds: BoundsType) -> None:
        """Checks if domain is not empty.

        :param bounds: The bounds to be checked.

        :raises ValueError: If bounds list is empty.
        :raises ValueError: If volume of hypercube defined by the bounds is emtpy.

        """
        if len(bounds) == 0:
            raise ValueError("Length of bound list must be > 0; empty list found.")

        for bounds_d in bounds:
            lb_d, ub_d = bounds_d
            if lb_d >= ub_d:
                raise ValueError(
                    "Upper box bound must be larger than lower bound. Found a pair containing ({}, "
                    "{}).".format(lb_d, ub_d)
                )

    def convert_to_list_of_continuous_parameters(self) -> List[ContinuousParameter]:
        """Converts the box bounds into a list of :class:`ContinuousParameter` objects.

        :return: The continuous parameters (one for each dimension of the box).
        """
        continuous_parameters = []
        for i, bounds_d in enumerate(self._bounds):
            lb_d, ub_d = bounds_d
            name_d = self.name + "_" + str(i)
            param = ContinuousParameter(name=name_d, min_value=lb_d, max_value=ub_d)
            continuous_parameters.append(param)
        return continuous_parameters
